# -*- coding: utf-8 -*-
from __future__ import annotations

import math
from typing import Final, Optional

from qtpy.QtGui import QValidator
from qtpy.QtWidgets import QDoubleSpinBox, QWidget

__all__ = ['FloatSpinBox']


class FloatSpinBox(QDoubleSpinBox):
    MODES: Final[list[str]] = ['auto', 'fixed', 'scientific']

    def __init__(self, parent: Optional[QWidget] = None) -> None:
        super().__init__(parent)
        self._mode: str = 'auto'
        self._decimals: int = 2
        super().setDecimals(1000)

    @property
    def mode(self) -> str:
        return self._mode

    @mode.setter
    def mode(self, new_mode: str) -> None:
        if new_mode not in self.MODES:
            raise ValueError(f'Invalid mode: {new_mode}')
        self._mode = new_mode

    def valueFromText(self, text: str) -> float:
        return float(text.strip().split()[0])

    def textFromValue(self, v: float) -> str:
        if self._mode == 'auto':
            if abs(v) < self.singleStep() and v != 0.0:
                return f'{v:.{self._decimals}e}'
            else:
                return f'{v:.{self._decimals}f}'
        elif self._mode == 'fixed':
            return f'{v:.{self._decimals}f}'
        elif self._mode == 'scientific':
            return f'{v:.{self._decimals}e}'
        else:
            raise RuntimeError(f'Unknown mode {self._mode}')

    def validate(self, text: str, pos: int) -> tuple[QValidator.State, str, int]:
        try:
            self.valueFromText(text)
        except (ValueError, TypeError):
            return QValidator.State.Invalid, text, pos
        else:
            return QValidator.State.Acceptable, text, pos

    def fixup(self, text: str) -> str:
        for word in text.split():
            try:
                float(word)
            except ValueError:
                continue
            else:
                return word
        return ''

    def stepBy(self, steps: int) -> None:
        if self.value() != 0.0:
            exp: int = round(math.floor(math.log10(abs(self.value()))))
            self.setValue(self.value() + self.singleStep() * steps * 10.0 ** exp)
        else:
            self.setValue(self.singleStep() * steps)

    def decimals(self) -> int:
        return self._decimals

    def setDecimals(self, new_value: int) -> None:
        if new_value >= 0:
            self._decimals = new_value
        else:
            raise ValueError(f'Invalid decimals value: {new_value}')



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
from __future__ import annotations

import sys

if sys.version_info < (3, 8):
    message = ('The Python version ' + '.'.join(map(str, sys.version_info[:3])) + ' is not supported.\n' +
               'Use Python 3.8 or newer.')
    try:
        import tkinter
    except ImportError:
        input(message)  # wait for the user to see the text
    else:
        print(message, file=sys.stderr)

        import tkinter.messagebox

        _root = tkinter.Tk()
        _root.withdraw()
        tkinter.messagebox.showerror(title='Outdated Python', message=message)
        _root.destroy()

    exit(1)

if __name__ == '__main__':

    try:
        from pycatsearch import main
    except ImportError:
        __author__ = 'StSav012'
        __original_name__ = 'py''cat''search'

        try:
            from updater import update_with_pip

            update_with_pip(__original_name__)

            from pycatsearch import main
        except ImportError:
            from updater import update_with_pip, update_from_github, update_with_git

            update_with_git() or update_from_github(__author__, __original_name__)

            from src.pycatsearch import main
    main()



# -*- coding: utf-8 -*-
from __future__ import annotations

from typing import Collection

from qtpy.QtCore import QModelIndex, Qt, Slot
from qtpy.QtWidgets import (QDialog, QDialogButtonBox, QFormLayout, QLabel, QListWidget, QListWidgetItem, QVBoxLayout,
                            QWidget)

from .selectable_label import SelectableLabel
from .url_label import URLLabel
from ..catalog import Catalog
from ..utils import *

__all__ = ['SubstanceInfoSelector', 'SubstanceInfo']


class SubstanceInfoSelector(QDialog):
    def __init__(self, catalog: Catalog, entry_ids: Collection[int],
                 allow_html: bool = True, inchi_key_search_url_template: str = '',
                 parent: QWidget | None = None) -> None:
        super().__init__(parent)
        self._catalog: Catalog = catalog
        self._inchi_key_search_url_template: str = inchi_key_search_url_template
        self.setModal(True)
        self.setWindowTitle(self.tr('Select Substance'))
        if parent is not None:
            self.setWindowIcon(parent.windowIcon())
        layout: QVBoxLayout = QVBoxLayout(self)
        self._list_box: QListWidget = QListWidget(self)
        self._list_box.itemSelectionChanged.connect(self._on_list_selection_changed)
        self._list_box.doubleClicked.connect(self._on_list_double_clicked)
        layout.addWidget(self._list_box)
        entry_ids = set(entry_ids)
        for entry in catalog.catalog:
            if not entry_ids:
                break  # nothing to search for
            if entry[ID] in entry_ids:
                entry_ids.discard(entry[ID])  # don't search for the ID again
                # don't specify the parent here: https://t.me/qtforpython/20950
                item: QListWidgetItem = QListWidgetItem()
                item.setData(Qt.ItemDataRole.ToolTipRole, str(entry[SPECIES_TAG]))
                item.setData(Qt.ItemDataRole.UserRole, entry[ID])
                self._list_box.addItem(item)
                self._list_box.setItemWidget(item, QLabel(best_name(entry, allow_html=allow_html)))
        self._buttons: QDialogButtonBox = QDialogButtonBox(
            (QDialogButtonBox.StandardButton.Ok | QDialogButtonBox.StandardButton.Close), self)
        self._buttons.button(QDialogButtonBox.StandardButton.Ok).setEnabled(False)
        self._buttons.accepted.connect(self._on_accept)
        self._buttons.rejected.connect(self.reject)
        layout.addWidget(self._buttons)

    @Slot()
    def _on_list_selection_changed(self) -> None:
        self._buttons.button(QDialogButtonBox.StandardButton.Ok).setEnabled(bool(self._list_box.selectedIndexes()))

    @Slot(QModelIndex)
    def _on_list_double_clicked(self, index: QModelIndex) -> None:
        self.hide()
        item: QListWidgetItem = self._list_box.item(index.row())
        syn: SubstanceInfo = SubstanceInfo(self._catalog, item.data(Qt.ItemDataRole.UserRole),
                                           inchi_key_search_url_template=self._inchi_key_search_url_template,
                                           parent=self.parent())
        syn.exec()
        self.accept()

    @Slot()
    def _on_accept(self) -> None:
        selected_items: list[QListWidgetItem] = self._list_box.selectedItems()
        if not selected_items:
            return
        self.hide()
        item: QListWidgetItem = selected_items.pop()
        syn: SubstanceInfo = SubstanceInfo(self._catalog, item.data(Qt.ItemDataRole.UserRole),
                                           inchi_key_search_url_template=self._inchi_key_search_url_template,
                                           parent=self.parent())
        syn.exec()
        self.accept()


class SubstanceInfo(QDialog):
    """ A simple dialog that displays the information about a substance from the loaded catalog """

    def __init__(self, catalog: Catalog, entry_id: int, inchi_key_search_url_template: str = '',
                 parent: QWidget | None = None) -> None:
        super().__init__(parent)
        self.setModal(True)
        self.setWindowTitle(self.tr('Substance Info'))
        if parent is not None:
            self.setWindowIcon(parent.windowIcon())
        layout: QFormLayout = QFormLayout(self)
        label: SelectableLabel
        for entry in catalog.catalog:
            if entry[ID] == entry_id:
                for key in entry:
                    if key == LINES:
                        continue
                    elif key == ID:
                        label = URLLabel(
                            url=f'https://cdms.astro.uni-koeln.de/cdms/portal/catalog/{entry[key]}/',
                            text=f'{entry[key]}',
                            parent=self)
                        label.setOpenExternalLinks(True)
                    elif key == STATE_HTML:
                        label = SelectableLabel(chem_html(str(entry[key])), self)
                    elif key == INCHI_KEY and inchi_key_search_url_template:
                        label = URLLabel(
                            url=inchi_key_search_url_template.format(InChIKey=entry[key]),
                            text=entry[key],
                            parent=self)
                        label.setOpenExternalLinks(True)
                    else:
                        label = SelectableLabel(str(entry[key]), self)
                    layout.addRow(self.tr(HUMAN_READABLE[key]), label)
                break
        buttons: QDialogButtonBox = QDialogButtonBox(QDialogButtonBox.StandardButton.Close, self)
        buttons.rejected.connect(self.reject)
        layout.addWidget(buttons)



# coding=utf-8
from __future__ import annotations

import argparse
import platform
import sys
from datetime import datetime, timedelta, timezone

from .catalog import Catalog

__author__ = 'StSav012'
__original_name__ = 'py''cat''search'

try:
    from ._version import __version__
except ImportError:
    __version__ = ''


def _version_tuple(version_string: str) -> tuple[int | str, ...]:
    result: tuple[int | str, ...] = tuple()
    part: str
    for part in version_string.split('.'):
        try:
            result += (int(part),)
        except ValueError:
            # follow `pkg_resources` version 0.6a9: remove dashes to sort letters after digits
            result += (part.replace('-', ''),)
    return result


def _warn_about_outdated_package(package_name: str, package_version: str, release_time: datetime) -> None:
    """ Display a warning about an outdated package a year after the package released """
    if datetime.now(tz=timezone.utc) - release_time > timedelta(days=366):
        import tkinter.messagebox
        tkinter.messagebox.showwarning(
            title='Package Outdated',
            message=f'Please update {package_name} package to {package_version} or newer')


def _make_old_qt_compatible_again() -> None:
    from qtpy import QT6, PYSIDE2
    from qtpy.QtCore import QLibraryInfo, Qt
    from qtpy.QtWidgets import QApplication, QDialog

    def to_iso_format(s: str) -> str:
        if sys.version_info < (3, 11):
            import re
            from typing import Callable

            if s.endswith('Z'):
                # '2011-11-04T00:05:23Z'
                s = s[:-1] + '+00:00'

            def from_iso_datetime(m: re.Match[str]) -> str:
                groups: dict[str, str] = m.groupdict('')
                date: str = f"{m['year']}-{m['month']}-{m['day']}"
                time: str = \
                    f"{groups['hour']:0>2}:{groups['minute']:0>2}:{groups['second']:0>2}.{groups['fraction']:0<6}"
                return date + 'T' + time + groups['offset']

            def from_iso_calendar(m: re.Match[str]) -> str:
                from datetime import date

                groups: dict[str, str] = m.groupdict('')
                date: str = \
                    date.fromisocalendar(year=int(m['year']), week=int(m['week']), day=int(m['dof'])).isoformat()
                time: str = \
                    f"{groups['hour']:0>2}:{groups['minute']:0>2}:{groups['second']:0>2}.{groups['fraction']:0<6}"
                return date + 'T' + time + groups['offset']

            patterns: dict[str, Callable[[re.Match[str]], str]] = {
                # '20111104', '20111104T000523283'
                r'(?P<year>\d{4})(?P<month>\d{2})(?P<day>\d{2})'
                r'(.(?P<hour>\d{2})(?P<minute>\d{2})(?P<second>\d{2})(?P<fraction>\d+)?)?'
                r'(?P<offset>[+\-].+)?': from_iso_datetime,
                # '2011-11-04', '2011-11-04T00:05:23.283', '2011-11-04T00:05:23.283+00:00'
                r'(?P<year>\d{4})-(?P<month>\d{1,2})-(?P<day>\d{1,2})'
                r'(.(?P<hour>\d{1,2}):(?P<minute>\d{1,2}):(?P<second>\d{1,2})(\.(?P<fraction>\d+))?)?'
                r'(?P<offset>[+\-].+)?': from_iso_datetime,
                # '2011-W01-2T00:05:23.283'
                r'(?P<year>\d{4})-W(?P<week>\d{1,2})-(?P<dof>\d{1,2})'
                r'(.(?P<hour>\d{1,2}):(?P<minute>\d{1,2}):(?P<second>\d{1,2})(\.(?P<fraction>\d+))?)?'
                r'(?P<offset>[+\-].+)?': from_iso_calendar,
                # '2011W0102T000523283'
                r'(?P<year>\d{4})-W(?P<week>\d{2})-(?P<dof>\d{2})'
                r'(.(?P<hour>\d{1,2})(?P<minute>\d{1,2})(?P<second>\d{1,2})(?P<fraction>\d+)?)?'
                r'(?P<offset>[+\-].+)?': from_iso_calendar,
            }
            match: re.Match[str] | None
            for p in patterns:
                match = re.fullmatch(p, s)
                if match is not None:
                    s = patterns[p](match)
                    break

        return s

    if PYSIDE2:
        QApplication.exec = QApplication.exec_
        QDialog.exec = QDialog.exec_

    if not QT6:
        QApplication.setAttribute(Qt.ApplicationAttribute.AA_EnableHighDpiScaling)
        QApplication.setAttribute(Qt.ApplicationAttribute.AA_UseHighDpiPixmaps)

    from qtpy import __version__

    if _version_tuple(__version__) < _version_tuple('2.3.1'):
        _warn_about_outdated_package(package_name='QtPy', package_version='2.3.1',
                                     release_time=datetime.fromisoformat(to_iso_format('2023-03-28T23:06:05Z')))
        if QT6:
            QLibraryInfo.LibraryLocation = QLibraryInfo.LibraryPath
    if _version_tuple(__version__) < _version_tuple('2.4.0'):
        # 2.4.0 is not released yet, so no warning until there is the release time
        if not QT6:
            QLibraryInfo.path = lambda *args, **kwargs: QLibraryInfo.location(*args, **kwargs)
            QLibraryInfo.LibraryPath = QLibraryInfo.LibraryLocation


def main() -> int:
    ap: argparse.ArgumentParser = argparse.ArgumentParser(
        allow_abbrev=True,
        description='Yet another implementation of JPL and CDMS spectroscopy catalogs offline search.\n'
                    f'Find more at https://github.com/{__author__}/{__original_name__}.')
    ap.add_argument('catalog', type=str, help='the catalog location to load',
                    nargs=argparse.ZERO_OR_MORE)
    ap_group = ap.add_argument_group(title='Search options',
                                     description='If any of the following arguments specified, a search conducted.')
    ap_group.add_argument('-f''min', '--min-frequency', type=float, help='the lower frequency [MHz] to take')
    ap_group.add_argument('-f''max', '--max-frequency', type=float, help='the upper frequency [MHz] to take')
    ap_group.add_argument('-i''min', '--min-intensity', type=float,
                          help='the minimal intensity [log10(nm²×MHz)] to take')
    ap_group.add_argument('-i''max', '--max-intensity', type=float,
                          help='the maximal intensity [log10(nm²×MHz)] to take')
    ap_group.add_argument('-T', '--temperature', type=float,
                          help='the temperature [K] to calculate the line intensity at, use the catalog intensity if not set')
    ap_group.add_argument('-t', '--tag', '--species-tag', type=int, dest='species_tag',
                          help='a number to match the `species''tag` field')
    ap_group.add_argument('-n', '--any-name-or-formula', type=str,
                          help='a string to match any field used by `any_name` and `any_formula` options')
    ap_group.add_argument('-a', '--anything', type=str, help='a string to match any field at all')
    ap_group.add_argument('--any-name', type=str, help='a string to match the `trivial''name` or the `name` field')
    ap_group.add_argument('--any-formula', type=str,
                          help='a string to match the `structural''formula`, `molecule''symbol`, '
                               '`stoichiometric''formula`, or `isotopolog` field')
    ap_group.add_argument('--InChI-key', '--inchi-key', type=str, dest='inchi_key',
                          help='a string to match the `inchikey` field, '
                               'which contains the IUPAC International Chemical Identifier (InChI™)')
    ap_group.add_argument('--trivial-name', type=str, help='a string to match the `trivial''name` field')
    ap_group.add_argument('--structural-formula', type=str,
                          help='a string to match the `structural''formula` field')
    ap_group.add_argument('--name', type=str, help='a string to match the `name` field')
    ap_group.add_argument('--stoichiometric-formula', type=str,
                          help='a string to match the `stoichiometric''formula` field')
    ap_group.add_argument('--isotopolog', type=str, help='a string to match the `isotopolog` field')
    ap_group.add_argument('--state', type=str, help='a string to match the `state` or `state_html` field')
    ap_group.add_argument('--dof', '--degrees_of_freedom', type=int, dest='degrees_of_freedom',
                          help='0 for atoms, 2 for linear molecules, and 3 for nonlinear molecules')
    ap_group.add_argument('--timeout', type=float, help='maximum time span for the filtering to take')

    args: argparse.Namespace = ap.parse_intermixed_args()

    arg_names: list[str] = ['min_frequency', 'max_frequency', 'min_intensity', 'max_intensity',
                            'temperature', 'species_tag', 'any_name_or_formula', 'anything', 'any_name', 'any_formula',
                            'inchi_key', 'trivial_name', 'structural_formula', 'name', 'stoichiometric_formula',
                            'isotopolog', 'state', 'degrees_of_freedom', 'timeout']
    search_args: dict[str, str | float | int] = dict((arg, getattr(args, arg)) for arg in arg_names
                                                     if getattr(args, arg) is not None)
    if search_args:
        c: Catalog = Catalog(*args.catalog)
        c.print(**search_args)
        return 0

    return main_gui()


def main_gui() -> int:
    ap: argparse.ArgumentParser = argparse.ArgumentParser(
        description='Yet another implementation of JPL and CDMS spectroscopy catalogs offline search.\n'
                    f'Find more at https://github.com/{__author__}/{__original_name__}.')
    ap.add_argument('catalog', type=str, help='the catalog location to load',
                    nargs=argparse.ZERO_OR_MORE)

    try:
        _make_old_qt_compatible_again()

        from . import gui
    except Exception as ex:
        import traceback
        from contextlib import suppress

        traceback.print_exc()

        error_message: str
        if isinstance(ex, SyntaxError):
            error_message = ('Python ' + platform.python_version() + ' is not supported.\n' +
                             'Get a newer Python!')
        elif isinstance(ex, ImportError):
            error_message = ('Module ' + repr(ex.name) +
                             ' is either missing from the system or cannot be loaded for another reason.\n' +
                             'Try to install or reinstall it.')
        else:
            error_message = str(ex)

        try:
            import tkinter
            import tkinter.messagebox
        except ModuleNotFoundError:
            input(error_message)
        else:
            print(error_message, file=sys.stderr)

            root: tkinter.Tk = tkinter.Tk()
            root.withdraw()
            if isinstance(ex, SyntaxError):
                tkinter.messagebox.showerror(title='Syntax Error', message=error_message)
            elif isinstance(ex, ImportError):
                tkinter.messagebox.showerror(title='Package Missing', message=error_message)
            else:
                tkinter.messagebox.showerror(title='Error', message=error_message)
            root.destroy()
        return 1
    else:
        return gui.run()


def download() -> None:
    from . import downloader

    downloader.download()


def async_download() -> None:
    from . import async_downloader

    async_downloader.download()



# -*- coding: utf-8 -*-
import sys
from contextlib import suppress
from typing import Any

from qtpy.QtGui import QIcon
from qtpy.QtWidgets import QApplication

from .ui import UI
from ..catalog import Catalog

__all__ = ['UI', 'icon', 'run']


def icon(*qta_name: str, **qta_specs: Any) -> QIcon:
    if qta_name:
        with suppress(ImportError, Exception):
            from qtawesome import icon

            return icon(*qta_name, **qta_specs)  # might raise an `Exception` if the icon is not in the font

    return QIcon()


def run() -> int:
    app: QApplication = QApplication(sys.argv)
    window: UI = UI(Catalog(*sys.argv[1:]))
    window.show()
    return app.exec()



# -*- coding: utf-8 -*-
from __future__ import annotations

import math
from typing import Any, Callable, Final, final

from qtpy.QtCore import (QAbstractTableModel, QByteArray, QItemSelection, QMimeData, QModelIndex, QPersistentModelIndex,
                         QPoint, QPointF, QRect, QSize, Qt, Slot)
from qtpy.QtGui import (QAbstractTextDocumentLayout, QClipboard, QCloseEvent, QCursor, QIcon, QPainter, QPixmap,
                        QScreen, QTextDocument)
from qtpy.QtWidgets import (QAbstractItemView, QAbstractSpinBox, QApplication, QDoubleSpinBox, QFormLayout, QHeaderView,
                            QMainWindow, QMessageBox, QPushButton, QSplitter, QStatusBar, QStyle, QStyleOptionViewItem,
                            QStyledItemDelegate, QTableView, QVBoxLayout, QWidget)
from qtpy.compat import getopenfilenames

from .catalog_info import CatalogInfo
from .download_dialog import DownloadDialog
from .float_spinbox import FloatSpinBox
from .frequency_box import FrequencyBox
from .menu_bar import MenuBar
from .preferences import Preferences
from .settings import Settings
from .substance_info import SubstanceInfo
from .substances_box import SubstancesBox
from .. import __version__
from ..catalog import Catalog, CatalogEntryType
from ..utils import *

__all__ = ['UI']


def copy_to_clipboard(text: str, text_type: Qt.TextFormat | str = Qt.TextFormat.PlainText) -> None:
    clipboard: QClipboard = QApplication.clipboard()
    if not text:
        return
    mime_data: QMimeData = QMimeData()
    if isinstance(text_type, str):
        mime_data.setData(text_type, text.encode())
    elif text_type == Qt.TextFormat.RichText:
        mime_data.setHtml(wrap_in_html(text))
        mime_data.setText(remove_html(text))
    else:
        mime_data.setText(text)
    clipboard.setMimeData(mime_data, QClipboard.Mode.Clipboard)


def substitute(fmt: str, *args: Any) -> str:
    res: str = fmt
    for index, value in enumerate(args):
        res = res.replace(f'{{{index}}}', str(value))
    return res


class HTMLDelegate(QStyledItemDelegate):
    @staticmethod
    def anchorAt(html: str, point: QPoint | QPointF) -> str:
        doc: QTextDocument = QTextDocument()
        doc.setHtml(html)
        text_layout: QAbstractTextDocumentLayout = doc.documentLayout()
        return text_layout.anchorAt(point)

    def paint(self, painter: QPainter, option: QStyleOptionViewItem, index: QModelIndex | QPersistentModelIndex) -> None:
        self.initStyleOption(option, index)
        style: QStyle
        if option.widget:
            style = option.widget.style()
        else:
            style = QApplication.style()
        doc: QTextDocument = QTextDocument()
        doc.setHtml(option.text)
        option.text = ''
        style.drawControl(QStyle.ControlElement.CE_ItemViewItem, option, painter)
        ctx: QAbstractTextDocumentLayout.PaintContext = QAbstractTextDocumentLayout.PaintContext()
        text_rect: QRect = style.subElementRect(QStyle.SubElement.SE_ItemViewItemText, option)
        painter.save()
        painter.translate(text_rect.topLeft())
        painter.setClipRect(text_rect.translated(-text_rect.topLeft()))
        painter.translate(0, 0.5 * (option.rect.height() - doc.size().height()))
        doc.documentLayout().draw(painter, ctx)
        painter.restore()

    def sizeHint(self, option: QStyleOptionViewItem | None, index: QModelIndex | QPersistentModelIndex) -> QSize:
        options: QStyleOptionViewItem = QStyleOptionViewItem(option)
        self.initStyleOption(options, index)
        doc: QTextDocument = QTextDocument()
        doc.setHtml(options.text)
        doc.setTextWidth(options.rect.width())
        return QSize(round(doc.idealWidth()), round(doc.size().height()))


class LinesListModel(QAbstractTableModel):
    ROW_BATCH_COUNT: Final[int] = 5

    class DataType:
        __slots__ = ['id', 'name',
                     'frequency_str', 'frequency',
                     'intensity_str', 'intensity',
                     'lower_state_energy_str', 'lower_state_energy']

        def __init__(self,
                     species_id: int, name: str,
                     frequency_str: str, frequency: float,
                     intensity_str: str, intensity: float,
                     lower_state_energy_str: str, lower_state_energy: float) -> None:
            self.id: int = species_id
            self.name: str = name
            self.frequency_str: str = frequency_str
            self.frequency: float = frequency
            self.intensity_str: str = intensity_str
            self.intensity: float = intensity
            self.lower_state_energy_str: str = lower_state_energy_str
            self.lower_state_energy: float = lower_state_energy

        def __eq__(self, other: LinesListModel.DataType) -> int:
            if not isinstance(other, LinesListModel.DataType):
                return NotImplemented
            return (self.id == other.id
                    and self.frequency == other.frequency
                    and self.intensity == other.intensity
                    and self.lower_state_energy == other.lower_state_energy)

        def __hash__(self) -> int:
            return hash(self.id) ^ hash(self.frequency) ^ hash(self.lower_state_energy)

    def __init__(self, settings: Settings, parent: QWidget | None = None) -> None:
        super().__init__(parent)
        self._settings: Settings = settings
        self._entries: list[CatalogEntryType] = []
        self._data: list[LinesListModel.DataType] = []
        self._rows_loaded: int = LinesListModel.ROW_BATCH_COUNT

        unit_format: Final[str] = self.tr('{0} [{1}]', 'unit format')
        self._header: Final[list[str]] = [
            self.tr('Substance'),
            substitute(unit_format, self.tr("Frequency"), self._settings.frequency_unit_str),
            substitute(unit_format, self.tr("Intensity"), self._settings.intensity_unit_str),
            substitute(unit_format, self.tr("Lower state energy"), self._settings.energy_unit_str),
        ]

    def update_units(self) -> None:
        unit_format: Final[str] = self.tr('{0} [{1}]', 'unit format')
        self._header[1] = substitute(unit_format, self.tr("Frequency"), self._settings.frequency_unit_str)
        self._header[2] = substitute(unit_format, self.tr("Intensity"), self._settings.intensity_unit_str)
        self._header[3] = substitute(unit_format, self.tr("Lower state energy"), self._settings.energy_unit_str)

    def rowCount(self, parent: QModelIndex | QPersistentModelIndex = ...) -> int:
        return min(len(self._data), self._rows_loaded)

    def columnCount(self, parent: QModelIndex | QPersistentModelIndex = ...) -> int:
        return len(self._header)

    def data(self, index: QModelIndex | QPersistentModelIndex, role: int = Qt.ItemDataRole.DisplayRole) -> str | None:
        if index.isValid():
            if role == Qt.ItemDataRole.DisplayRole:
                item: LinesListModel.DataType = self._data[index.row()]
                column_index: int = index.column()
                if column_index == 0:
                    return item.name
                if column_index == 1:
                    return item.frequency_str
                if column_index == 2:
                    return item.intensity_str
                if column_index == 3:
                    return item.lower_state_energy_str
        return None

    def row(self, row_index: int) -> LinesListModel.DataType:
        return self._data[row_index]

    def headerData(self, col: int, orientation: Qt.Orientation, role: int = ...) -> str | None:
        if orientation == Qt.Orientation.Horizontal and role == Qt.ItemDataRole.DisplayRole:
            return self._header[col]
        return None

    def setHeaderData(self, section: int, orientation: Qt.Orientation, value: str, role: int = ...) -> bool:
        if (orientation == Qt.Orientation.Horizontal
                and role == Qt.ItemDataRole.DisplayRole
                and 0 <= section < len(self._header)):
            self._header[section] = value
            return True
        return False

    def clear(self) -> None:
        self.set_entries([])

    def set_entries(self, new_data: list[CatalogEntryType]) -> None:
        from_mhz: Callable[[float], float] = self._settings.from_mhz
        from_log10_sq_nm_mhz: Callable[[float], float] = self._settings.from_log10_sq_nm_mhz
        from_rec_cm: Callable[[float], float] = self._settings.from_rec_cm
        frequency_suffix: int = self._settings.frequency_unit
        precision: int = [4, 7, 8, 8][frequency_suffix]

        def frequency_str(frequency: float) -> tuple[str, float]:
            frequency = from_mhz(frequency)
            return f'{frequency:.{precision}f}', frequency

        def intensity_str(intensity: float) -> tuple[str, float]:
            intensity = from_log10_sq_nm_mhz(intensity)
            if intensity == 0.0:
                return '0', intensity
            elif abs(intensity) < 0.1:
                return f'{intensity:.4e}', intensity
            else:
                return f'{intensity:.4f}', intensity

        def lower_state_energy_str(lower_state_energy: float) -> tuple[str, float]:
            lower_state_energy = from_rec_cm(lower_state_energy)
            if lower_state_energy == 0.0:
                return '0', lower_state_energy
            elif abs(lower_state_energy) < 0.1:
                return f'{lower_state_energy:.4e}', lower_state_energy
            else:
                return f'{lower_state_energy:.4f}', lower_state_energy

        self.beginResetModel()
        unique_entries: list[CatalogEntryType] = []
        non_unique_indices: set[int] = set()
        unique: bool
        all_unique: bool = True  # unless the opposite is proven
        for i in range(len(new_data)):
            if i in non_unique_indices:
                continue
            unique = True
            for j in range(i + 1, len(new_data)):
                if j in non_unique_indices:
                    continue
                if new_data[i] == new_data[j]:
                    non_unique_indices.add(j)
                    unique = False
                    all_unique = False
                    break
            if unique and not all_unique:
                unique_entries.append(new_data[i])
        if all_unique:
            self._entries = new_data.copy()
        else:
            self._entries = unique_entries
        entry: CatalogEntryType
        rich_text_in_formulas: bool = self._settings.rich_text_in_formulas
        self._data = list(set(
            LinesListModel.DataType(
                entry[ID],
                best_name(entry, rich_text_in_formulas),
                *frequency_str(line[FREQUENCY]),
                *intensity_str(line[INTENSITY]),
                *lower_state_energy_str(line[LOWER_STATE_ENERGY]),
            )
            for entry in self._entries
            for line in entry[LINES]
        ))
        self._rows_loaded = LinesListModel.ROW_BATCH_COUNT
        self.endResetModel()

    def sort(self, column: int, order: Qt.SortOrder = Qt.SortOrder.AscendingOrder) -> None:
        self.beginResetModel()
        key = {
            0: (lambda l: (l.name, l.frequency, l.intensity, l.lower_state_energy)),
            1: (lambda l: (l.frequency, l.intensity, l.name, l.lower_state_energy)),
            2: (lambda l: (l.intensity, l.frequency, l.name, l.lower_state_energy)),
            3: (lambda l: (l.lower_state_energy, l.intensity, l.frequency, l.name))
        }[column]
        self._data.sort(key=key, reverse=bool(order != Qt.SortOrder.AscendingOrder))
        self.endResetModel()

    def canFetchMore(self, index: QModelIndex | QPersistentModelIndex = QModelIndex()) -> bool:
        return len(self._data) > self._rows_loaded

    def fetchMore(self, index: QModelIndex | QPersistentModelIndex = QModelIndex()) -> None:
        # https://sateeshkumarb.wordpress.com/2012/04/01/paginated-display-of-table-data-in-pyqt/
        remainder: int = len(self._data) - self._rows_loaded
        items_to_fetch: int = min(remainder, LinesListModel.ROW_BATCH_COUNT)
        self.beginInsertRows(QModelIndex(), self._rows_loaded, self._rows_loaded + items_to_fetch - 1)
        self._rows_loaded += items_to_fetch
        self.endInsertRows()


@final
class UI(QMainWindow):
    def __init__(self, catalog: Catalog,
                 parent: QWidget | None = None) -> None:
        super().__init__(parent)
        self.catalog: Catalog = catalog
        self.settings: Settings = Settings('SavSoft', 'CatSearch', self)

        self._central_widget: QSplitter = QSplitter(Qt.Orientation.Vertical, self)
        self._top_matter: QSplitter = QSplitter(Qt.Orientation.Horizontal, self._central_widget)
        self._right_matter: QWidget = QWidget(self._central_widget)

        self.spin_intensity: FloatSpinBox = FloatSpinBox(self._central_widget)
        self.spin_temperature: QDoubleSpinBox = QDoubleSpinBox(self._central_widget)

        self.box_substance: SubstancesBox = SubstancesBox(self.catalog, self.settings, self._central_widget)
        self.box_frequency: FrequencyBox = FrequencyBox(self.settings, self._central_widget)
        self.button_search: QPushButton = QPushButton(self._central_widget)

        self.results_model: LinesListModel = LinesListModel(self.settings, self)
        self.results_table: QTableView = QTableView(self._central_widget)

        self.menu_bar: MenuBar = MenuBar(self)

        self.status_bar: QStatusBar = QStatusBar(self)

        def setup_ui() -> None:
            from . import icon  # import locally to avoid a circular import

            # https://ru.stackoverflow.com/a/1032610
            window_icon: QPixmap = QPixmap()
            window_icon.loadFromData(b'''\
            <svg height="64" width="64" version="1.1">
            <path stroke-linejoin="round" d="m6.722 8.432c-9.05 9.648-6.022 27.23 6.048 33.04 6.269 3.614 13.88 \
            3.1 20-0.1664l20 20c2.013 2.013 5.256 2.013 7.27 0l1.259-1.259c2.013-2.013 2.013-5.256 \
            0-7.27l-19.83-19.83c1.094-1.948 1.868-4.095 2.211-6.403 3.06-13.5-9.72-27.22-23.4-25.12-4.74 \
            0.53-9.28 2.72-12.64 6.104-0.321 0.294-0.626 0.597-0.918 0.908zm8.015 6.192c4.978-5.372 14.79-3.878 17.96 \
            2.714 3.655 6.341-0.6611 15.28-7.902 16.36-7.14 1.62-14.4-5.14-13.29-12.38 0.2822-2.51 1.441-4.907 \
            3.231-6.689z" stroke="#000" stroke-width="2.4" fill="#fff"/>
            </svg>''')
            self.setWindowIcon(QIcon(window_icon))

            if __version__:
                self.setWindowTitle(self.tr('PyCatSearch (version {0})').format(__version__))
            else:
                self.setWindowTitle(self.tr('PyCatSearch'))
            self.setCentralWidget(self._central_widget)

            layout_right: QVBoxLayout = QVBoxLayout()
            layout_options: QFormLayout = QFormLayout()

            self.results_table.setModel(self.results_model)
            self.results_table.setItemDelegateForColumn(0, HTMLDelegate())
            self.results_table.setMouseTracking(True)
            self.results_table.setContextMenuPolicy(Qt.ContextMenuPolicy.CustomContextMenu)
            self.results_table.setEditTriggers(QAbstractItemView.EditTrigger.NoEditTriggers)
            self.results_table.setDropIndicatorShown(False)
            self.results_table.setDragDropOverwriteMode(False)
            self.results_table.setSelectionBehavior(QAbstractItemView.SelectionBehavior.SelectRows)
            self.results_table.setCornerButtonEnabled(False)
            self.results_table.setSortingEnabled(True)
            self.results_table.setAlternatingRowColors(True)
            self.results_table.horizontalHeader().setDefaultSectionSize(180)
            self.results_table.horizontalHeader().setHighlightSections(False)
            self.results_table.horizontalHeader().setSectionResizeMode(0, QHeaderView.ResizeMode.Stretch)
            self.results_table.horizontalHeader().setSectionResizeMode(1, QHeaderView.ResizeMode.ResizeToContents)
            self.results_table.horizontalHeader().setSectionResizeMode(2, QHeaderView.ResizeMode.ResizeToContents)
            self.results_table.horizontalHeader().setSectionResizeMode(3, QHeaderView.ResizeMode.ResizeToContents)
            self.results_table.horizontalHeader().setSectionsMovable(True)
            self.results_table.verticalHeader().setVisible(False)
            self.results_table.verticalHeader().setHighlightSections(False)

            # substance selection
            self._top_matter.addWidget(self.box_substance)

            # frequency limits
            layout_right.addWidget(self.box_frequency, 1)

            self.spin_intensity.setAlignment(Qt.AlignmentFlag.AlignRight
                                             | Qt.AlignmentFlag.AlignTrailing
                                             | Qt.AlignmentFlag.AlignVCenter)
            self.spin_intensity.setButtonSymbols(QAbstractSpinBox.ButtonSymbols.NoButtons)
            self.spin_intensity.setDecimals(2)
            self.spin_intensity.setMinimum(-math.inf)
            self.spin_intensity.setMaximum(math.inf)
            self.spin_intensity.setSingleStep(0.1)
            self.spin_intensity.setValue(-6.54)
            self.spin_intensity.setStatusTip(self.spin_intensity.tr('Limit shown spectral lines'))
            layout_options.addRow(layout_options.tr('Minimal Intensity:'), self.spin_intensity)
            self.spin_temperature.setAlignment(Qt.AlignmentFlag.AlignRight
                                               | Qt.AlignmentFlag.AlignTrailing
                                               | Qt.AlignmentFlag.AlignVCenter)
            self.spin_temperature.setButtonSymbols(QAbstractSpinBox.ButtonSymbols.NoButtons)
            self.spin_temperature.setMaximum(999.99)
            self.spin_temperature.setValue(300.0)
            self.spin_temperature.setStatusTip(self.spin_temperature.tr('Temperature to calculate intensity'))
            self.spin_temperature.setSuffix(self.spin_temperature.tr(' K'))
            layout_options.addRow(layout_options.tr('Temperature:'), self.spin_temperature)
            layout_right.addLayout(layout_options, 0)

            self.button_search.setText(self.button_search.tr('Show'))
            layout_right.addWidget(self.button_search, 0)

            self._right_matter.setLayout(layout_right)
            self._top_matter.addWidget(self._right_matter)
            self._top_matter.setStretchFactor(0, 1)
            self._top_matter.setChildrenCollapsible(False)

            self._central_widget.addWidget(self._top_matter)
            self._central_widget.addWidget(self.results_table)
            self._central_widget.setStretchFactor(1, 1)
            self._central_widget.setChildrenCollapsible(False)

            self.setMenuBar(self.menu_bar)
            self.setStatusBar(self.status_bar)

            self.button_search.setShortcut('Ctrl+Return')

            self.button_search.setIcon(icon('mdi6.magnify'))

            self.adjustSize()

        setup_ui()

        self.temperature: float = 300.0  # [K]
        self.minimal_intensity: float = -math.inf  # [log10(nm²×MHz)]

        self.button_search.setDisabled(self.catalog.is_empty)

        self.preferences_dialog: Preferences = Preferences(self.settings, self)

        self.results_shown: bool = False

        self.preset_table()

        self.load_settings()

        self.results_table.customContextMenuRequested.connect(self._on_table_context_menu_requested)
        self.results_table.selectionModel().selectionChanged.connect(self._on_table_item_selection_changed)
        self.results_table.doubleClicked.connect(self._on_action_substance_info_triggered)
        self.spin_intensity.valueChanged.connect(self._on_spin_intensity_changed)
        self.spin_temperature.valueChanged.connect(self._on_spin_temperature_changed)
        self.button_search.clicked.connect(self._on_button_search_clicked)
        self.menu_bar.action_load.triggered.connect(self._on_action_load_triggered)
        self.menu_bar.action_quit.triggered.connect(self._on_action_quit_triggered)
        self.menu_bar.action_check_updates.triggered.connect(self._on_action_check_updates_triggered)
        self.menu_bar.action_about_catalogs.triggered.connect(self._on_action_about_catalogs_triggered)
        self.menu_bar.action_about.triggered.connect(self._on_action_about_triggered)
        self.menu_bar.action_about_qt.triggered.connect(self._on_action_about_qt_triggered)
        self.menu_bar.action_download_catalog.triggered.connect(self._on_action_download_catalog_triggered)
        self.menu_bar.action_preferences.triggered.connect(self._on_action_preferences_triggered)
        self.menu_bar.action_copy.triggered.connect(self._on_action_copy_triggered)
        self.menu_bar.action_select_all.triggered.connect(self._on_action_select_all_triggered)
        self.menu_bar.action_reload.triggered.connect(self._on_action_reload_triggered)
        self.menu_bar.action_copy_current.triggered.connect(self._on_action_copy_current_triggered)
        self.menu_bar.action_copy_name.triggered.connect(self._on_action_copy_name_triggered)
        self.menu_bar.action_copy_frequency.triggered.connect(self._on_action_copy_frequency_triggered)
        self.menu_bar.action_copy_intensity.triggered.connect(self._on_action_copy_intensity_triggered)
        self.menu_bar.action_copy_lower_state_energy.triggered.connect(
            self._on_action_copy_lower_state_energy_triggered)
        self.menu_bar.action_show_substance.toggled.connect(self._on_action_show_substance_toggled)
        self.menu_bar.action_show_frequency.toggled.connect(self._on_action_show_frequency_toggled)
        self.menu_bar.action_show_intensity.toggled.connect(self._on_action_show_intensity_toggled)
        self.menu_bar.action_show_lower_state_energy.toggled.connect(self._on_action_show_lower_state_energy_toggled)
        self.menu_bar.action_substance_info.triggered.connect(self._on_action_substance_info_triggered)
        self.menu_bar.action_clear.triggered.connect(self._on_action_clear_triggered)

        if not self.catalog.is_empty:
            self.box_frequency.set_frequency_limits(self.catalog.min_frequency, self.catalog.max_frequency)

        if self.settings.check_updates:
            _latest_release: ReleaseInfo = latest_release()
            if (_latest_release
                    and _latest_release.version != self.settings.ignored_version
                    and _latest_release.version > __version__):
                res: QMessageBox.StandardButton = QMessageBox.question(
                    self, self.tr('Release Info'),
                    self.tr('Version {release.version} published {release.pub_date} is available. '
                            'Would you like to get the update? '
                            'The app will try to restart.').format(release=_latest_release),
                    QMessageBox.StandardButton.Yes | QMessageBox.StandardButton.No | QMessageBox.StandardButton.Ignore)
                if res == QMessageBox.StandardButton.Yes:
                    update_with_pip()
                elif res == QMessageBox.StandardButton.Ignore:
                    self.settings.ignored_version = _latest_release.version

    def closeEvent(self, event: QCloseEvent) -> None:
        self.save_settings()
        event.accept()

    def load_catalog(self, *catalog_file_names: str) -> bool:
        self.setDisabled(True)
        last_cursor: QCursor = self.cursor()
        self.setCursor(Qt.CursorShape.WaitCursor)
        self.repaint()
        self.catalog = Catalog(*catalog_file_names)
        self.box_substance.catalog = self.catalog
        self.setCursor(last_cursor)
        self.setEnabled(True)
        self.button_search.setDisabled(self.catalog.is_empty)
        if not self.catalog.is_empty:
            self.box_frequency.set_frequency_limits(self.catalog.min_frequency, self.catalog.max_frequency)
        return not self.catalog.is_empty

    @Slot(float)
    def _on_spin_temperature_changed(self, arg1: float) -> None:
        self.temperature = self.settings.to_k(arg1)
        self.fill_table()

    @Slot(float)
    def _on_spin_intensity_changed(self, arg1: float) -> None:
        self.minimal_intensity = self.settings.to_log10_sq_nm_mhz(arg1)
        if self.results_shown:
            self.fill_table()

    @Slot(QPoint)
    def _on_table_context_menu_requested(self, pos: QPoint) -> None:
        self.menu_bar.menu_edit.popup(self.results_table.viewport().mapToGlobal(pos))

    @Slot(QItemSelection, QItemSelection)
    def _on_table_item_selection_changed(self, _selected: QItemSelection, _deselected: QItemSelection) -> None:
        self.menu_bar.action_copy.setEnabled(bool(self.results_table.selectionModel().selectedRows()))
        self.menu_bar.action_substance_info.setEnabled(bool(self.results_table.selectionModel().selectedRows()))

    def get_open_file_names(self, formats: dict[tuple[str, ...], str],
                            caption: str = '', directory: str = '') -> tuple[list[str], str]:

        def join_file_dialog_formats(_formats: dict[tuple[str, ...], str]) -> str:
            f: tuple[str, ...]
            all_supported_extensions: list[str] = []
            for f in _formats.keys():
                all_supported_extensions.extend(ensure_prefix(_f, '*') for _f in f)
            format_lines: list[str] = [''.join((
                self.tr('All supported', 'file type'),
                '(',
                ' '.join(ensure_prefix(_f, '*') for _f in all_supported_extensions),
                ')'))]
            n: str
            for f, n in _formats.items():
                format_lines.append(''.join((n, '(', ' '.join(ensure_prefix(_f, '*') for _f in f), ')')))
            format_lines.append(self.tr('All files', 'file type') + '(* *.*)')
            return ';;'.join(format_lines)

        filename: list[str]
        _filter: str
        filename, _filter = getopenfilenames(self,
                                             caption=caption,
                                             filters=join_file_dialog_formats(formats),
                                             basedir=directory)
        return filename, _filter

    @Slot()
    def _on_action_load_triggered(self) -> None:
        self.status_bar.showMessage(self.tr('Select a catalog file to load.'))
        new_catalog_file_names: list[str]
        _formats: dict[tuple[str, ...], str] = {
            (*all_cases('.json.gz'), *all_cases('.json.bz2'),
             *all_cases('.json.xz'), *all_cases('.json.lzma')): self.tr('Compressed JSON', 'file type'),
            tuple(all_cases('.json')): self.tr('JSON', 'file type'),
        }
        new_catalog_file_names, _ = self.get_open_file_names(formats=_formats,
                                                             caption=self.tr('Load Catalog'),
                                                             directory=(*self.catalog.sources, '')[0])

        if new_catalog_file_names:
            self.status_bar.showMessage(self.tr('Loading...'))
            if self.load_catalog(*new_catalog_file_names):
                self.status_bar.showMessage(self.tr('Catalogs loaded.'))
            else:
                self.status_bar.showMessage(self.tr('Failed to load a catalog.'))

        else:
            self.status_bar.clearMessage()

    @Slot()
    def _on_action_reload_triggered(self) -> None:
        if self.catalog.sources:
            self.status_bar.showMessage(self.tr('Loading...'))
            if self.load_catalog(*self.catalog.sources):
                self.status_bar.showMessage(self.tr('Catalogs loaded.'))
            else:
                self.status_bar.showMessage(self.tr('Failed to load a catalog.'))
        else:
            self.status_bar.clearMessage()

    def stringify_selection_html(self) -> str:
        """
        Convert selected rows to string for copying as rich text
        :return: the rich text representation of the selected table lines
        """
        if not self.results_table.selectionModel().selectedRows():
            return ''

        units: list[str] = [
            '',
            self.settings.frequency_unit_str,
            self.settings.intensity_unit_str,
            self.settings.energy_unit_str,
        ]
        with_units: bool = self.settings.with_units

        def format_value(value: Any, unit: str) -> str:
            return (self.tr('{value} {unit}', 'format value in html').format(value=value, unit=unit)
                    if with_units and unit
                    else self.tr('{value}', 'format value in html').format(value=value))

        columns_order: list[int] = [self.results_table.horizontalHeader().logicalIndex(_c)
                                    for _c, _a in zip(range(self.results_table.horizontalHeader().count()),
                                                      self.menu_bar.menu_columns.actions(),
                                                      strict=True)
                                    if _a.isChecked()]
        text: list[str] = ['<table>']
        values: list[str]
        index: QModelIndex
        for index in self.results_table.selectionModel().selectedRows():
            row: LinesListModel.DataType = self.results_model.row(index.row())
            values = [
                format_value(_v, _u)
                for _u, _v, _a in zip(units,
                                      (row.name, row.frequency, row.intensity, row.lower_state_energy),
                                      self.menu_bar.menu_columns.actions(),
                                      strict=True)
                if _a.isChecked()
            ]
            text.append(
                '<tr><td>' +
                f'</td>{self.settings.csv_separator}<td>'.join(values[_c] for _c in columns_order) +
                '</td></tr>'
            )
        text.append('</table>')
        return self.settings.line_end.join(text)

    @Slot()
    def _on_action_download_catalog_triggered(self) -> None:
        downloader: DownloadDialog = DownloadDialog(
            frequency_limits=(self.catalog.min_frequency, self.catalog.max_frequency),
            parent=self)
        downloader.exec()

    @Slot()
    def _on_action_preferences_triggered(self) -> None:
        self.preferences_dialog.exec()
        self.fill_parameters()
        if self.results_model.rowCount():
            self.preset_table()
            self.fill_table()
        else:
            self.preset_table()

    @Slot()
    def _on_action_quit_triggered(self) -> None:
        self.close()

    @Slot()
    def _on_action_clear_triggered(self) -> None:
        self.results_model.clear()
        self.preset_table()

    def copy_selected_items(self, col: int) -> None:
        if col >= self.results_model.columnCount():
            return

        def html_list(lines: list[str]) -> str:
            return '<ul><li>' + f'</li>{self.settings.line_end}<li>'.join(lines) + '</li></ul>'

        text_to_copy: list[str] = []
        index: QModelIndex
        for index in (self.results_table.selectionModel().selectedRows(col)
                      or [self.results_table.selectionModel().currentIndex()]):
            if index.isValid():
                text_to_copy.append(self.results_model.data(index))
        if not text_to_copy:
            return
        if col == 0:
            copy_to_clipboard(html_list(text_to_copy), Qt.TextFormat.RichText)
        else:
            copy_to_clipboard(self.settings.line_end.join(text_to_copy), Qt.TextFormat.PlainText)

    @Slot()
    def _on_action_copy_current_triggered(self) -> None:
        self.copy_selected_items(self.results_table.selectionModel().currentIndex().column())

    @Slot()
    def _on_action_copy_name_triggered(self) -> None:
        self.copy_selected_items(0)

    @Slot()
    def _on_action_copy_frequency_triggered(self) -> None:
        self.copy_selected_items(1)

    @Slot()
    def _on_action_copy_intensity_triggered(self) -> None:
        self.copy_selected_items(2)

    @Slot()
    def _on_action_copy_lower_state_energy_triggered(self) -> None:
        self.copy_selected_items(3)

    @Slot()
    def _on_action_copy_triggered(self) -> None:
        copy_to_clipboard(self.stringify_selection_html(), Qt.TextFormat.RichText)

    @Slot()
    def _on_action_select_all_triggered(self) -> None:
        self.results_table.selectAll()

    @Slot()
    def _on_action_substance_info_triggered(self) -> None:
        if self.results_table.selectionModel().selectedRows():
            syn: SubstanceInfo = SubstanceInfo(
                self.catalog,
                self.results_model.row(self.results_table.selectionModel().selectedRows()[0].row()).id,
                inchi_key_search_url_template=self.settings.inchi_key_search_url_template,
                parent=self)
            syn.exec()

    def toggle_results_table_column_visibility(self, column: int, is_visible: bool) -> None:
        if is_visible != self.results_table.isColumnHidden(column):
            return
        if is_visible:
            self.results_table.showColumn(column)
        else:
            self.results_table.hideColumn(column)

    @Slot(bool)
    def _on_action_show_substance_toggled(self, is_checked: bool) -> None:
        self.toggle_results_table_column_visibility(0, is_checked)

    @Slot(bool)
    def _on_action_show_frequency_toggled(self, is_checked: bool) -> None:
        self.toggle_results_table_column_visibility(1, is_checked)

    @Slot(bool)
    def _on_action_show_intensity_toggled(self, is_checked: bool) -> None:
        self.toggle_results_table_column_visibility(2, is_checked)

    @Slot(bool)
    def _on_action_show_lower_state_energy_toggled(self, is_checked: bool) -> None:
        self.toggle_results_table_column_visibility(3, is_checked)

    @Slot()
    def _on_action_check_updates_triggered(self) -> None:
        _latest_release: ReleaseInfo = latest_release()
        if not _latest_release:
            QMessageBox.warning(self, self.tr('Release Info'), self.tr('Update check failed.'))
        elif _latest_release.version > __version__:
            res: QMessageBox.StandardButton = QMessageBox.question(
                self, self.tr('Release Info'),
                self.tr('Version {release.version} published {release.pub_date} is available. '
                        'Would you like to get the update? '
                        'The app will try to restart.').format(release=_latest_release),
                QMessageBox.StandardButton.Yes | QMessageBox.StandardButton.No | QMessageBox.StandardButton.Ignore)
            if res == QMessageBox.StandardButton.Yes:
                update_with_pip()
            elif res == QMessageBox.StandardButton.Ignore:
                self.settings.ignored_version = _latest_release.version
        else:
            QMessageBox.information(self, self.tr('Release Info'), self.tr('You are using the latest version.'))

    @Slot()
    def _on_action_about_catalogs_triggered(self) -> None:
        if self.catalog:
            ci: CatalogInfo = CatalogInfo(self.catalog, self)
            ci.exec()
        else:
            QMessageBox.information(self, self.tr('Catalog Info'), self.tr('No catalogs loaded'))

    @Slot()
    def _on_action_about_triggered(self) -> None:
        QMessageBox.about(self,
                          self.tr("About CatSearch"),
                          "<html><p>"
                          + self.tr("CatSearch is a means of searching through spectroscopy lines catalogs. "
                                    "It's an offline application.")
                          + "</p><p>"
                          + self.tr("It relies on the data stored in JSON files.")
                          + "</p><p>"
                          + self.tr("One can write his own catalogs as well as convert data from "
                                    "<a href='https://spec.jpl.nasa.gov/'>JPL</a> and "
                                    "<a href='https://astro.uni-koeln.de/'>CDMS</a> spectroscopy databases "
                                    "freely available in the Internet or other sources.")
                          + "</p><p>"
                          + self.tr("Both plain text JSON and GZip compressed JSON are supported.")
                          + "</p><p>"
                          + self.tr('See {0} for more info.')
                          .format('<a href="https://github.com/StSav012/pycatsearch/blob/master/README.md">{0}</a>')
                          .format(self.tr('readme'))
                          + "</p><br><p>"
                          + self.tr("CatSearch is licensed under the {0}.")
                          .format("<a href='https://www.gnu.org/copyleft/lesser.html'>{0}</a>"
                                  .format(self.tr("GNU LGPL version 3")))
                          + "</p><p>"
                          + self.tr("The source code is available on {0}.").format(
                              "<a href='https://github.com/StSav012/pycatsearch'>GitHub</a>")
                          + "</p></html>")

    @Slot()
    def _on_action_about_qt_triggered(self) -> None:
        QMessageBox.aboutQt(self)

    def load_settings(self) -> None:
        self.settings.beginGroup('search')
        catalog_file_names: list[str] = []
        for i in range(self.settings.beginReadArray('catalogFiles')):
            self.settings.setArrayIndex(i)
            path: str = self.settings.value('path', '', str)
            if path:
                catalog_file_names.append(path)
        self.settings.endArray()
        if not catalog_file_names:
            catalog_file_names = ['catalog.json.gz', 'catalog.json']
        self.temperature = self.settings.value('temperature', self.spin_temperature.value(), float)
        self.minimal_intensity = self.settings.value('intensity', self.spin_intensity.value(), float)
        self.settings.endGroup()
        self.settings.beginGroup('displayedColumns')
        self.menu_bar.action_show_substance.setChecked(self.settings.value('substance', True, bool))
        self.toggle_results_table_column_visibility(0, self.menu_bar.action_show_substance.isChecked())
        self.menu_bar.action_show_frequency.setChecked(self.settings.value('frequency', True, bool))
        self.toggle_results_table_column_visibility(1, self.menu_bar.action_show_frequency.isChecked())
        self.menu_bar.action_show_intensity.setChecked(self.settings.value('intensity', True, bool))
        self.toggle_results_table_column_visibility(2, self.menu_bar.action_show_intensity.isChecked())
        self.menu_bar.action_show_lower_state_energy.setChecked(self.settings.value('lowerStateEnergy', False, bool))
        self.toggle_results_table_column_visibility(3, self.menu_bar.action_show_lower_state_energy.isChecked())
        self.results_table.horizontalHeader().restoreState(self.settings.value('state', QByteArray()))
        self.results_table.horizontalHeader().restoreGeometry(self.settings.value('geometry', QByteArray()))
        self.settings.endGroup()
        self.settings.beginGroup('window')
        screens: list[QScreen] = QApplication.screens()
        if screens:
            self.move(round(0.5 * (screens[0].size().width() - self.size().width())),
                      round(0.5 * (screens[0].size().height() - self.size().height())))  # Fallback: Center the window
        self.restoreGeometry(self.settings.value('geometry', QByteArray()))
        self.restoreState(self.settings.value('state', QByteArray()))
        self._top_matter.restoreGeometry(self.settings.value('verticalSplitterGeometry', QByteArray()))
        self._top_matter.restoreState(self.settings.value('verticalSplitterState', QByteArray()))
        self._central_widget.restoreGeometry(self.settings.value('horizontalSplitterGeometry', QByteArray()))
        self._central_widget.restoreState(self.settings.value('horizontalSplitterState', QByteArray()))
        self.settings.endGroup()
        self.fill_parameters()

        if self.settings.load_last_catalogs:
            self.load_catalog(*catalog_file_names)

    def save_settings(self) -> None:
        self.settings.beginGroup('search')
        self.settings.beginWriteArray('catalogFiles', len(self.catalog.sources))
        for i, s in enumerate(self.catalog.sources):
            self.settings.setArrayIndex(i)
            self.settings.setValue('path', s)
        self.settings.endArray()
        self.settings.setValue('temperature', self.temperature)
        self.settings.setValue('intensity', self.minimal_intensity)
        self.settings.endGroup()
        self.settings.beginGroup('displayedColumns')
        self.settings.setValue('substance', self.menu_bar.action_show_substance.isChecked())
        self.settings.setValue('frequency', self.menu_bar.action_show_frequency.isChecked())
        self.settings.setValue('intensity', self.menu_bar.action_show_intensity.isChecked())
        self.settings.setValue('lowerStateEnergy', self.menu_bar.action_show_lower_state_energy.isChecked())
        self.settings.setValue('geometry', self.results_table.horizontalHeader().saveGeometry())
        self.settings.setValue('state', self.results_table.horizontalHeader().saveState())
        self.settings.endGroup()
        self.settings.beginGroup('window')
        self.settings.setValue('geometry', self.saveGeometry())
        self.settings.setValue('state', self.saveState())
        self.settings.setValue('verticalSplitterGeometry', self._top_matter.saveGeometry())
        self.settings.setValue('verticalSplitterState', self._top_matter.saveState())
        self.settings.setValue('horizontalSplitterGeometry', self._central_widget.saveGeometry())
        self.settings.setValue('horizontalSplitterState', self._central_widget.saveState())
        self.settings.endGroup()
        self.box_substance.save_settings()
        self.box_frequency.save_settings()
        self.settings.sync()

    def preset_table(self) -> None:
        self.results_shown = False
        self.results_table.clearSelection()
        self.menu_bar.action_copy.setDisabled(True)
        self.menu_bar.action_substance_info.setDisabled(True)
        self.menu_bar.action_select_all.setDisabled(True)
        self.menu_bar.action_clear.setDisabled(True)
        self.menu_bar.menu_copy_only.setDisabled(True)
        self.results_model.update_units()
        self.update()

    def fill_parameters(self) -> None:
        # frequency
        if not self.catalog.is_empty:
            self.box_frequency.set_frequency_limits(self.catalog.min_frequency, self.catalog.max_frequency)
        self.box_frequency.fill_parameters()

        # intensity
        self.spin_intensity.setSuffix(' ' + self.settings.intensity_unit_str)
        self.spin_intensity.setValue(self.settings.from_log10_sq_nm_mhz(self.minimal_intensity))

        # temperature
        temperature_suffix: int = self.settings.temperature_unit
        self.spin_temperature.setSuffix(' ' + self.settings.TEMPERATURE_UNITS[temperature_suffix])
        if temperature_suffix == 0:  # K
            self.spin_temperature.setValue(self.temperature)
            self.spin_temperature.setMinimum(0.0)
        elif temperature_suffix == 1:  # °C
            self.spin_temperature.setMinimum(-273.15)
            self.spin_temperature.setValue(self.settings.from_k(self.temperature))
        else:
            raise IndexError('Wrong temperature unit index', temperature_suffix)

    def fill_table(self) -> None:
        self.preset_table()
        self.results_table.setSortingEnabled(False)

        entries: list[dict[str, int | str | list[dict[str, float]]]] = \
            (sum(
                (
                    self.catalog.filter(min_frequency=self.box_frequency.min_frequency,
                                        max_frequency=self.box_frequency.max_frequency,
                                        min_intensity=self.minimal_intensity,
                                        anything=name,
                                        temperature=self.temperature,
                                        timeout=self.settings.timeout)
                    for name in self.box_substance.selected_substances
                ),
                []
            ) if self.box_substance.selected_substances and self.box_substance.isChecked()
             else self.catalog.filter(min_frequency=self.box_frequency.min_frequency,
                                      max_frequency=self.box_frequency.max_frequency,
                                      min_intensity=self.minimal_intensity,
                                      temperature=self.temperature,
                                      timeout=self.settings.timeout))
        self.results_model.set_entries(entries)

        self.results_table.setSortingEnabled(True)
        self.menu_bar.action_select_all.setEnabled(bool(entries))
        self.menu_bar.action_clear.setEnabled(bool(entries))
        self.menu_bar.menu_copy_only.setEnabled(bool(entries))
        self.results_shown = True

    @Slot()
    def _on_button_search_clicked(self) -> None:
        self.status_bar.showMessage(self.tr('Searching...'))
        self.setDisabled(True)
        last_cursor: QCursor = self.cursor()
        self.setCursor(Qt.CursorShape.WaitCursor)
        self.repaint()
        self.box_substance.update_selected_substances()
        self.fill_table()
        self.setCursor(last_cursor)
        self.setEnabled(True)
        self.status_bar.showMessage(self.tr('Ready.'))



# -*- coding: utf-8 -*-
from __future__ import annotations

import asyncio
import logging
import random
from contextlib import suppress
from math import inf
from queue import Empty, Queue
from threading import Thread
from typing import Any, Final, Mapping, cast
from urllib.error import HTTPError
from urllib.parse import urlencode

import aiohttp
import aiohttp.client_exceptions

try:
    import orjson as json
except ImportError:
    import json

from .catalog_entry import CatalogEntry
from .utils import *

__all__ = ['Downloader', 'get_catalog', 'save_catalog', 'download']

logger: logging.Logger = logging.getLogger('async_downloader')


class Downloader(Thread):
    def __init__(self,
                 frequency_limits: tuple[float, float] = (-inf, inf),
                 state_queue: Queue[tuple[int, int]] | None = None) -> None:
        super().__init__()
        self._state_queue: Queue[tuple[int, int]] | None = state_queue
        self._frequency_limits: tuple[float, float] = frequency_limits
        self._catalog: list[dict[str, int | str | list[dict[str, float]]]] = []

        self._run: bool = False
        self._tasks: list[asyncio.Task] = []

    def __del__(self) -> None:
        self._run = False

    @property
    def catalog(self) -> list[dict[str, int | str | list[dict[str, float]]]]:
        return self._catalog.copy()

    def join(self, timeout: float | None = ...) -> None:
        self._run = False
        self.cancel()
        loop: asyncio.AbstractEventLoop = asyncio.get_event_loop()
        loop.stop()
        loop.close()
        super().join(timeout=timeout)

    def cancel(self) -> None:
        for task in self._tasks:
            if not task.done():
                task.cancel()

    def run(self) -> None:
        self._run = True

        async def async_get_catalog() -> list[dict[str, int | str | list[dict[str, float]]]]:

            session: aiohttp.ClientSession
            async with aiohttp.ClientSession(timeout=aiohttp.ClientTimeout(),  # disable timeout checks
                                             trust_env=True) as session:

                async def get(url: str, headers: Mapping[str, str] | None = None) -> bytes:
                    response: aiohttp.ClientResponse
                    while self._run:
                        try:
                            async with session.get(url, headers=headers) as response:
                                return await response.read()
                        except asyncio.exceptions.CancelledError as ex:
                            if self._run:
                                logger.error(str(ex), exc_info=ex)
                        except aiohttp.client_exceptions.ServerDisconnectedError as ex:
                            logger.warning(f'{url}: {str(ex.message)}')
                        except aiohttp.client_exceptions.ClientConnectorError as ex:
                            logger.warning(f'{str(ex.args[1])} to {url}')
                        except aiohttp.client_exceptions.ClientOSError as ex:
                            logger.warning(f'{url}: {str(ex)}')
                        except aiohttp.client_exceptions.ClientError as ex:
                            logger.error(f'{url}: {str(ex)}', exc_info=ex)
                        with suppress(asyncio.exceptions.CancelledError):
                            await asyncio.sleep(random.random())
                    return bytes()

                async def post(url: str, data: dict[str, Any], headers: Mapping[str, str] | None = None) -> str:
                    async with session.post(url, data=urlencode(data).encode(), headers=headers) as response:
                        return (await response.read()).decode()

                async def get_species() -> list[dict[str, int | str]]:
                    def purge_null_data(entry: dict[str, None | int | str]) -> dict[str, int | str]:
                        key: str
                        value: None | int | str
                        return dict((key, value) for key, value in entry.items()
                                    if value is not None and value not in ('', 'None'))

                    def trim_strings(entry: dict[str, None | int | str]) -> dict[str, None | int | str]:
                        key: str
                        for key in entry:
                            if isinstance(entry[key], str):
                                entry[key] = cast(str, entry[key]).strip()
                        return entry

                    data: dict[str, int | str | list[dict[str, None | int | str]]] \
                        = json.loads(await post('https://cdms.astro.uni-koeln.de/cdms/portal/json_list/species/',
                                                {'database': -1},
                                                headers={'Content-Type': 'application/x-www-form-urlencoded'}))
                    if 'species' in data:
                        return [purge_null_data(trim_strings(s)) for s in data['species']]
                    else:
                        return []

                async def get_substance_catalog(species_entry: dict[str, int | str]) \
                        -> dict[str, int | str | list[dict[str, float]]]:
                    def entry_url(species_tag: int) -> str:
                        entry_filename: str = f'c{species_tag:06}.cat'

                        if entry_filename in ('c044009.cat', 'c044012.cat'):
                            return ''  # merged with c044004.cat — Brian J. Drouin
                        if species_tag % 1000 > 500:
                            return 'https://cdms.astro.uni-koeln.de/classic/entries/' + entry_filename
                        else:
                            return 'https://spec.jpl.nasa.gov/ftp/pub/catalog/' + entry_filename

                    if SPECIES_TAG not in species_entry:
                        # nothing to go on with
                        return dict()
                    fn: str = entry_url(species_tag=cast(int, species_entry[SPECIES_TAG]))
                    if not fn:  # no need to download a file for the species tag
                        return dict()
                    try:
                        lines = (await get(fn)).decode().splitlines()
                    except HTTPError as ex:
                        logger.error(fn, exc_info=ex)
                        return dict()
                    catalog_entries = [CatalogEntry(line) for line in lines]
                    if not catalog_entries:
                        if self._run:
                            logger.warning('no entries in the catalog')
                        return dict()
                    return {
                        **species_entry,
                        DEGREES_OF_FREEDOM: catalog_entries[0].degrees_of_freedom,
                        LINES: [_catalog_entry.to_dict()
                                for _catalog_entry in catalog_entries
                                if within(_catalog_entry.frequency, self._frequency_limits)]
                    }

                species: list[dict[str, int | str]] = await get_species()
                self._tasks = [asyncio.create_task(get_substance_catalog(_e)) for _e in species]
                catalog: list[dict[str, int | str | list[dict[str, float]]]] = []
                species_count: Final[int] = len(species)
                skipped_count: int = 0
                catalog_entry: dict[str, int | str | list[dict[str, float]]]
                future_entry_index: int
                future_entry: asyncio.Future[dict[str, int | str | list[dict[str, float]]]]
                for future_entry_index, future_entry in enumerate(asyncio.as_completed(self._tasks), start=1):
                    catalog_entry = await future_entry
                    if catalog_entry and LINES in catalog_entry and catalog_entry[LINES]:
                        catalog.append(catalog_entry)
                    else:
                        skipped_count += 1
                    if self._state_queue is not None:
                        self._state_queue.put((len(catalog), species_count - future_entry_index - skipped_count))

            if self._state_queue is not None:
                self._state_queue.put((len(catalog), 0))

            return catalog

        with suppress(RuntimeError):  # it might be “cannot schedule new futures after shutdown”
            self._catalog = asyncio.run(async_get_catalog())


def get_catalog(frequency_limits: tuple[float, float] = (-inf, inf)) \
        -> list[dict[str, int | str | list[dict[str, float]]]]:
    """
    Download the spectral lines catalog data

    :param tuple frequency_limits: the frequency range of the catalog entries to keep.
    :return: a list of the spectral lines catalog entries.
    """

    state_queue: Queue[tuple[int, int]] = Queue()
    downloader: Downloader = Downloader(frequency_limits=frequency_limits, state_queue=state_queue)
    downloader.start()
    while downloader.is_alive():
        cataloged_species: int
        not_yet_processed_species: int
        try:
            cataloged_species, not_yet_processed_species = state_queue.get(block=True, timeout=0.1)
        except Empty:
            continue
        else:
            logger.info(f'got {cataloged_species} entries, {not_yet_processed_species} left')

    return downloader.catalog


def save_catalog(filename: str,
                 frequency_limits: tuple[float, float] = (0, inf)) -> bool:
    """
    Download and save the spectral lines catalog data

    :param str filename: the name of the file to save the downloaded catalog to.
        If it ends with an unknown suffix, `'.json.gz'` is appended to it.
    :param tuple frequency_limits: the tuple of the maximal and the minimal frequencies of the lines being stored.
        All the lines outside the specified frequency range are omitted.
    """

    return save_catalog_to_file(filename=filename,
                                catalog=get_catalog(frequency_limits),
                                frequency_limits=frequency_limits)


def download() -> None:
    import argparse
    from datetime import datetime

    ap: argparse.ArgumentParser = argparse.ArgumentParser(
        allow_abbrev=True,
        description='Download JPL and CDMS spectroscopy catalogs for offline search.\n'
                    'Find more at https://github.com/StSav012/pycatsearch.')
    ap.add_argument('catalog', type=str, help='the catalog location to save into (required)')
    ap.add_argument('-f''min', '--min-frequency', type=float, help='the lower frequency [MHz] to take', default=-inf)
    ap.add_argument('-f''max', '--max-frequency', type=float, help='the upper frequency [MHz] to take', default=+inf)
    args: argparse.Namespace = ap.parse_intermixed_args()

    logging.basicConfig(level=logging.DEBUG)
    logger.info(f'started at {datetime.now()}')
    save_catalog(args.catalog, (args.min_frequency, args.max_frequency))
    logger.info(f'finished at {datetime.now()}')



# -*- coding: utf-8 -*-
from __future__ import annotations

from typing import Sequence

from qtpy.QtWidgets import QLabel, QListWidget, QVBoxLayout, QWidget

__all__ = ['TitledListWidget']


class TitledListWidget(QWidget):
    def __init__(self, parent: QWidget | None = None) -> None:
        super().__init__(parent)
        layout: QVBoxLayout = QVBoxLayout(self)
        self._title_label: QLabel = QLabel(self)
        layout.addWidget(self._title_label)
        self._list_widget: QListWidget = QListWidget(self)
        layout.addWidget(self._list_widget)
        layout.setContentsMargins(0, 0, 0, 0)
        layout.setStretch(0, 0)

    def setTitle(self, title: str) -> None:
        self._title_label.setText(title)

    def addItems(self, labels: Sequence[str]) -> None:
        self._list_widget.addItems(labels)



# -*- coding: utf-8 -*-
from __future__ import annotations

from threading import Thread
from typing import Any, Callable, Mapping, Sequence

from qtpy.QtCore import QCoreApplication, QSize, Qt
from qtpy.QtWidgets import QLabel, QVBoxLayout, QWidget

__all__ = ['WaitingScreen']


def _spinner(parent: QWidget | None = None) -> QWidget | None:
    from contextlib import suppress

    with suppress(ImportError, Exception):
        import qtawesome as qta

        spinner: qta.IconWidget = qta.IconWidget(parent=parent)
        spinner.setIconSize(QSize(*([spinner.fontMetrics().height() * 2] * 2)))
        # might raise an `Exception` if the icon is not in the font
        spinner.setIcon(qta.icon('mdi6.loading', animation=qta.Spin(spinner, interval=16, step=4)))
        spinner.setAlignment(Qt.AlignmentFlag.AlignCenter)
        return spinner

    return None


class WaitingScreen(QWidget):
    def __init__(self, parent: QWidget, label: str | QWidget,
                 target: Callable[[...], Any], args: Sequence[Any] = (), kwargs: Mapping[str, Any] | None = None,
                 margins: float | None = None) -> None:
        super().__init__(parent, Qt.WindowType.SplashScreen)

        self.setWindowModality(Qt.WindowModality.WindowModal)

        if isinstance(label, str):
            label = QLabel(label, self)
            label.setAlignment(Qt.AlignmentFlag.AlignCenter)
        layout: QVBoxLayout = QVBoxLayout(self)
        spinner: QWidget | None = _spinner(self)
        if spinner is not None:
            layout.addWidget(spinner)
        layout.addWidget(label)

        if margins is not None:
            layout.setContentsMargins(*([margins] * 4))

        self._target: Callable[[...], Any] = target
        self._args: Sequence[Any] = args
        self._kwargs: Mapping[str, Any] = kwargs or dict()
        self._thread: Thread | None = None

    @property
    def active(self) -> bool:
        return self._thread is not None and self._thread.is_alive()

    def exec(self) -> None:
        self._thread = Thread(
            target=self._target,
            args=self._args,
            kwargs=self._kwargs)
        self.show()
        self._thread.start()
        while self.active:
            QCoreApplication.processEvents()
        self._thread.join()
        self._thread = None
        self.hide()

    def stop(self) -> None:
        if self._thread is not None:
            self._thread.join(0.0)
        self._thread = None


if __name__ == '__main__':
    import sys

    from qtpy.QtWidgets import QApplication

    app: QApplication = QApplication(sys.argv)
    w: WaitingScreen = WaitingScreen(None, 'label', None)  # type: ignore
    w.hideEvent = lambda event: app.quit()
    w.show()
    app.exec()



# -*- coding: utf-8 -*-
from __future__ import annotations

from qtpy.QtCore import QModelIndex, Qt, Slot
from qtpy.QtWidgets import (QAbstractItemView, QAbstractScrollArea, QCheckBox, QGroupBox, QLineEdit, QListWidget,
                            QListWidgetItem, QPushButton, QVBoxLayout, QWidget)

from .settings import Settings
from .substance_info import SubstanceInfo, SubstanceInfoSelector
from ..catalog import Catalog
from ..utils import *

__all__ = ['SubstancesBox']


class SubstancesBox(QGroupBox):
    def __init__(self, catalog: Catalog, settings: Settings, parent: QWidget | None = None) -> None:
        from . import icon  # import locally to avoid a circular import

        super().__init__(parent)

        self._catalog: Catalog = catalog
        self._settings: Settings = settings
        self._selected_substances: set[str] = set()

        self._layout_substance: QVBoxLayout = QVBoxLayout(self)
        self._text_substance: QLineEdit = QLineEdit(self)
        self._list_substance: QListWidget = QListWidget(self)
        self._check_keep_selection: QCheckBox = QCheckBox(self)
        self._button_select_none: QPushButton = QPushButton(self)

        self.setCheckable(True)
        self.setTitle(self.tr('Search Only For…'))
        self._text_substance.setClearButtonEnabled(True)
        self._text_substance.setPlaceholderText(self._text_substance.tr('Filter'))
        self._layout_substance.addWidget(self._text_substance)
        self._list_substance.setEditTriggers(QAbstractItemView.EditTrigger.NoEditTriggers)
        self._list_substance.setDropIndicatorShown(False)
        self._list_substance.setAlternatingRowColors(True)
        self._list_substance.setSelectionMode(QAbstractItemView.SelectionMode.ExtendedSelection)
        self._list_substance.setSelectionBehavior(QAbstractItemView.SelectionBehavior.SelectRows)
        self._list_substance.setSortingEnabled(False)
        self._list_substance.setSizeAdjustPolicy(QAbstractScrollArea.SizeAdjustPolicy.AdjustToContents)
        self._layout_substance.addWidget(self._list_substance)
        self._check_keep_selection.setStatusTip(
            self._check_keep_selection.tr('Keep substances list selection through filter changes'))
        self._check_keep_selection.setText(self._check_keep_selection.tr('Persistent Selection'))
        self._layout_substance.addWidget(self._check_keep_selection)
        self._button_select_none.setStatusTip(self._button_select_none.tr('Clear substances list selection'))
        self._button_select_none.setText(self._button_select_none.tr('Select None'))
        self._layout_substance.addWidget(self._button_select_none)

        self._button_select_none.setIcon(icon('mdi6.checkbox-blank-off-outline'))

        self._text_substance.textChanged.connect(self._on_text_changed)
        self._check_keep_selection.toggled.connect(self._on_check_save_selection_toggled)
        self._button_select_none.clicked.connect(self._on_button_select_none_clicked)
        self._list_substance.doubleClicked.connect(self._on_list_substance_double_clicked)

        self.load_settings()

    def update_selected_substances(self) -> None:
        if self.isChecked():
            for i in range(self._list_substance.count()):
                item: QListWidgetItem = self._list_substance.item(i)
                if item.checkState() == Qt.CheckState.Checked:
                    self._selected_substances.add(item.text())
                else:
                    self._selected_substances.discard(item.text())
        else:
            self._selected_substances.clear()

    def filter_substances_list(self, filter_text: str) -> dict[str, set[int]]:
        list_items: dict[str, set[int]] = dict()
        plain_text_name: str
        if filter_text:
            filter_text_lowercase: str = filter_text.casefold()
            for name_key in (ISOTOPOLOG, NAME, STRUCTURAL_FORMULA,
                             STOICHIOMETRIC_FORMULA, TRIVIAL_NAME):
                for entry in self._catalog.catalog:
                    plain_text_name = remove_html(str(entry[name_key]))
                    if (name_key in entry
                            and (plain_text_name.startswith(filter_text)
                                 or (name_key in (NAME, TRIVIAL_NAME)
                                     and plain_text_name.casefold().startswith(filter_text_lowercase)))):
                        if plain_text_name not in list_items:
                            list_items[plain_text_name] = set()
                        list_items[plain_text_name].add(entry[ID])
            for name_key in (ISOTOPOLOG, NAME, STRUCTURAL_FORMULA,
                             STOICHIOMETRIC_FORMULA, TRIVIAL_NAME):
                for entry in self._catalog.catalog:
                    plain_text_name = remove_html(str(entry[name_key]))
                    if (name_key in entry
                            and (filter_text in plain_text_name
                                 or (name_key in (NAME, TRIVIAL_NAME)
                                     and filter_text_lowercase in plain_text_name.casefold()))):
                        if plain_text_name not in list_items:
                            list_items[plain_text_name] = set()
                        list_items[plain_text_name].add(entry[ID])
            # species tag suspected
            if filter_text.isdecimal():
                for entry in self._catalog.catalog:
                    plain_text_name = str(entry.get(SPECIES_TAG, ''))
                    if plain_text_name.startswith(filter_text):
                        if plain_text_name not in list_items:
                            list_items[plain_text_name] = set()
                        list_items[plain_text_name].add(entry[ID])
            # InChI Key match, see https://en.wikipedia.org/wiki/International_Chemical_Identifier#InChIKey
            if (len(filter_text) == 27
                    and filter_text[14] == '-' and filter_text[25] == '-'
                    and filter_text.count('-') == 2):
                for entry in self._catalog.catalog:
                    plain_text_name = str(entry.get(INCHI_KEY, ''))
                    if plain_text_name == filter_text:
                        if plain_text_name not in list_items:
                            list_items[plain_text_name] = set()
                        list_items[plain_text_name].add(entry[ID])
        else:
            for name_key in (ISOTOPOLOG, NAME, STRUCTURAL_FORMULA,
                             STOICHIOMETRIC_FORMULA, TRIVIAL_NAME):
                for entry in self._catalog.catalog:
                    plain_text_name = remove_html(str(entry[name_key]))
                    if plain_text_name not in list_items:
                        list_items[plain_text_name] = set()
                    list_items[plain_text_name].add(entry[ID])
            list_items = dict(sorted(list_items.items()))
        return list_items

    def fill_substances_list(self, filter_text: str | None = None) -> None:
        if not filter_text:
            filter_text = self._text_substance.text()

        self.update_selected_substances()
        self._list_substance.clear()

        text: str
        ids: set[int]
        for text, ids in self.filter_substances_list(filter_text).items():
            new_item: QListWidgetItem = QListWidgetItem(text)
            new_item.setData(Qt.ItemDataRole.UserRole, ids)
            new_item.setFlags(new_item.flags() | Qt.ItemFlag.ItemIsUserCheckable)
            new_item.setCheckState(Qt.CheckState.Checked
                                   if text in self._selected_substances
                                   else Qt.CheckState.Unchecked)
            self._list_substance.addItem(new_item)

    @Slot(str)
    def _on_text_changed(self, current_text: str) -> None:
        self.fill_substances_list(current_text)

    @Slot(bool)
    def _on_check_save_selection_toggled(self, new_state: bool) -> None:
        if not new_state:
            self._selected_substances.clear()
            self.update_selected_substances()

    @Slot()
    def _on_button_select_none_clicked(self) -> None:
        for i in range(self._list_substance.count()):
            self._list_substance.item(i).setCheckState(Qt.CheckState.Unchecked)
        self._selected_substances.clear()

    @Slot(QModelIndex)
    def _on_list_substance_double_clicked(self, index: QModelIndex) -> None:
        item: QListWidgetItem = self._list_substance.item(index.row())
        ids: set[int] = item.data(Qt.ItemDataRole.UserRole).copy()
        if len(ids) > 1:
            sis: SubstanceInfoSelector = SubstanceInfoSelector(
                self.catalog, ids,
                inchi_key_search_url_template=self._settings.inchi_key_search_url_template,
                parent=self)
            sis.exec()
        elif ids:  # if not empty
            syn: SubstanceInfo = SubstanceInfo(
                self.catalog, ids.pop(),
                inchi_key_search_url_template=self._settings.inchi_key_search_url_template,
                parent=self)
            syn.exec()

    def load_settings(self) -> None:
        self._settings.beginGroup('search')
        self._settings.beginGroup('selection')
        self._text_substance.setText(self._settings.value('filter', self._text_substance.text(), str))
        self._check_keep_selection.setChecked(self._settings.value('isPersistent', False, bool))
        self.setChecked(self._settings.value('enabled', self.isChecked(), bool))
        self._settings.endGroup()
        self._settings.endGroup()

    def save_settings(self) -> None:
        self._settings.beginGroup('search')
        self._settings.beginGroup('selection')
        self._settings.setValue('filter', self._text_substance.text())
        self._settings.setValue('isPersistent', self._check_keep_selection.isChecked())
        self._settings.setValue('enabled', self.isChecked())
        self._settings.endGroup()
        self._settings.endGroup()

    @property
    def catalog(self) -> Catalog:
        return self._catalog

    @catalog.setter
    def catalog(self, new_value: Catalog) -> None:
        self._catalog = new_value
        self.fill_substances_list()

    @property
    def selected_substances(self) -> set[str]:
        return self._selected_substances



# -*- coding: utf-8 -*-
from __future__ import annotations

import logging
import subprocess
import sys
import urllib.request
from contextlib import suppress
from datetime import datetime
from http.client import HTTPResponse
from pathlib import Path

__all__ = ['update_from_github', 'update_with_git', 'update_with_pip']

logger: logging.Logger = logging.getLogger('updater')


def from_iso_format(s: str) -> datetime:
    if sys.version_info < (3, 11):
        # NB: 'W' specifier is not fixed
        if s.endswith('Z'):  # '2011-11-04T00:05:23Z'
            s = s[:-1] + '+00:00'
        if s.isdigit() and len(s) == 8:  # '20111104'
            s = '-'.join((s[:4], s[4:6], s[6:]))
        elif s[:8].isdigit() and s[9:].isdigit() and len(s) >= 13:  # '20111104T000523'
            s = '-'.join((s[:4], s[4:6], s[6:8])) + s[8] + ':'.join((s[9:11], s[11:13], s[13:]))
    return datetime.fromisoformat(s)


def get_github_date(user: str, repo_name: str, branch: str = 'master') -> datetime | None:
    import json

    url: str = f'https://api.github.com/repos/{user}/{repo_name}/commits/{branch}'
    logger.debug(f'Requesting {url}')
    r: HTTPResponse = urllib.request.urlopen(url, timeout=1)
    logger.debug(f'Response code: {r.getcode()}')
    if r.getcode() != 200:
        logger.warning(f'Response code is not OK: {r.getcode()}')
        return None
    content: bytes = r.read()
    if not content:
        logger.warning(f'No data received from {url}')
        return None
    d: dict[str, str
                 | dict[str, bool | int | str]
                 | dict[str, int | str | dict[str, bool | str] | dict[str, str]]
                 | dict[str, int]
                 | list[dict[str, int | str]]
                 | list[dict[str, str]]] \
        = json.loads(content)
    if not isinstance(d, dict) or not d:
        logger.warning(f'Malformed JSON received: {d}')
        return None
    commit: dict[str, int | str | dict[str, bool | str] | dict[str, str]] = d.get('commit', dict())
    if not isinstance(commit, dict):
        logger.warning(f'Malformed commit info received: {commit}')
        return None
    committer: dict[str, str] = commit.get('committer', dict())
    if not isinstance(committer, dict) or 'date' not in committer:
        logger.warning(f'Malformed commit committer info received: {committer}')
        return None
    try:
        return from_iso_format(committer['date'])
    except ValueError:
        return None


def upgrade_files(code_directory: Path, user: str, repo_name: str, branch: str = 'master') -> bool:
    """ Replace the files in `code_directory` with the newer versions acquired from GitHub """

    import io
    import zipfile

    url: str = f'https://github.com/{user}/{repo_name}/archive/{branch}.zip'
    logger.debug(f'Requesting {url}')
    r: HTTPResponse = urllib.request.urlopen(url, timeout=1)
    logger.debug(f'Response code: {r.getcode()}')
    if r.getcode() != 200:
        logger.warning(f'Response code is not OK: {r.getcode()}')
        return False
    with zipfile.ZipFile(io.BytesIO(r.read())) as inner_zip:
        root: Path = Path(f'{repo_name}-{branch}/')
        member: zipfile.ZipInfo
        for member in inner_zip.infolist():
            logger.debug(f'Un-zipping {member.filename}')
            if member.is_dir():
                logger.debug('it is a directory')
                continue
            (code_directory / Path(member.filename).relative_to(root)).parent.mkdir(parents=True, exist_ok=True)
            (code_directory / Path(member.filename).relative_to(root)).write_bytes(inner_zip.read(member))
            logger.info(f'{(code_directory / Path(member.filename).relative_to(root))} written')

    return True


def update_with_git() -> bool:
    with suppress(Exception):
        code_directory: Path = Path(__file__).parent
        if (code_directory / '.git').exists():
            return subprocess.run(args=['git', 'pull'], capture_output=True).returncode == 0
    return False


def update_from_github(user: str, repo_name: str, branch: str = 'master') -> bool:
    with suppress(Exception):
        code_directory: Path = Path(__file__).parent
        version_path: Path = code_directory / 'src' / repo_name / '_version.py'

        github_date: datetime | None = get_github_date(user=user, repo_name=repo_name)
        if github_date is None:
            logger.warning('Failed to fetch the last commit date from GitHub')
            return False
        if version_path.exists() and datetime.fromtimestamp(version_path.stat().st_mtime) >= github_date:
            logger.info('Current files are up to date')
            return False

        if upgrade_files(code_directory=code_directory, user=user, repo_name=repo_name, branch=branch):
            # if everything went fine...
            version_path.parent.mkdir(exist_ok=True, parents=True)
            version_path.write_text(f'__version__ = version = "{github_date.isoformat()}"\n')
            logger.info(f'{github_date} written into {version_path}')
            return True
    return False


def parse_table(table_text: str) -> list[dict[str, str]]:
    text_lines: list[str] = table_text.splitlines()
    rules: list[str] = [line for line in text_lines if set(line) == set('- ')]
    if len(rules) != 1:
        raise RuntimeError('Failed to parse the table')
    if text_lines.index(rules[0]) != 1:
        raise RuntimeError('Failed to parse the table')
    cols: list[int] = [len(rule) for rule in rules[0].split()]
    titles: list[str] = []
    offset: int = 0
    for col in cols:
        titles.append(text_lines[0][offset:(offset + col)].strip())
        offset += col + 1
    data: list[dict[str, str]] = []
    for line_no in range(2, len(text_lines)):
        data.append(dict())
        offset = 0
        for col, title in zip(cols, titles):
            data[-1][title] = text_lines[line_no][offset:(offset + col)].strip()
            offset += col + 1
    return data


def update_package(package_name: str) -> tuple[str, str, int | None]:
    p: subprocess.CompletedProcess = subprocess.run(
        args=[sys.executable, '-m', 'pip', 'install', '-U', package_name],
        capture_output=True, text=True)
    return p.stdout, p.stderr, p.returncode


def update_packages() -> list[str]:
    priority_packages: list[str] = ['pip', 'setuptools', 'wheel']
    out: str
    err: str
    ret: int | None
    p: subprocess.CompletedProcess = subprocess.run(
        args=[sys.executable, '-m', 'pip', 'list', '--outdated'],
        capture_output=True, text=True)
    if p.returncode:
        return []
    outdated_packages: list[str] = [item['Package'] for item in parse_table(p.stdout)]
    updated_packages: list[str] = []
    try:
        for pp in priority_packages:
            if pp in outdated_packages:
                out, err, ret = update_package(pp)
                if ret:
                    return updated_packages
                outdated_packages.remove(pp)
                updated_packages.append(pp)
        for op in outdated_packages:
            update_package(op)
            updated_packages.append(op)
    finally:
        return updated_packages


def update_with_pip(package_name: str) -> bool:
    with suppress(Exception):
        if package_name not in update_packages():
            out, err, ret = update_package(package_name)
            return not ret
    return False


if __name__ == '__main__':
    import argparse

    ap: argparse.ArgumentParser = argparse.ArgumentParser(
        description='Fetch the code of a package from a GitHub repository')
    ap.add_argument('user', type=str, help='the owner of the GitHub repository')
    ap.add_argument('repo', type=str, help='the GitHub repository name')
    ap.add_argument('branch', type=str, help='the GitHub repository branch', default='master')

    args: argparse.Namespace = ap.parse_args()
    update_from_github(user=args.user, repo_name=args.repo, branch=args.branch)



# -*- coding: utf-8 -*-
from __future__ import annotations

import math

from .utils import *

__all__ = ['CatalogEntry']


class CatalogEntry:
    def __init__(self, spcat_line: str = '', *, frequency: float = math.nan, intensity: float = math.nan,
                 degrees_of_freedom: int = -1, lower_state_energy: float = math.nan) -> None:
        self.FREQ: float  # frequency, MHz, mandatory
        self.INT: float  # intensity, log10(nm²×MHz), mandatory
        self.DR: int  # degrees of freedom: 0 for atoms, 2 for linear molecules, and 3 for nonlinear molecules.
        self.ELO: float  # lower state energy relative to the ground state, 1/cm
        self.FREQ = frequency
        self.INT = intensity
        self.DR = degrees_of_freedom
        self.ELO = lower_state_energy
        if spcat_line:
            # FREQ         ERR     LGINT   DR ELO      GUP TAG   QNFMT QN'       QN"
            # F13     .4   F8 .4   F8 .4   I2F10  .4,  I3 I7     I4  6I2         6I2
            # FFFFFFFF.FFFFEEE.EEEE-II.IIIIDDEEEEE.EEEEGGG+TTTTTTQQQQ112233445566112233445566
            #      262.0870  0.0011-19.2529 2 5174.7303  4  180011335 1-132 2 2   1 132 2 3
            self.FREQ = float(spcat_line[:13])
            self.INT = float(spcat_line[21:29])
            self.DR = int(spcat_line[29:31])
            self.ELO = float(spcat_line[31:41])

    @property
    def frequency(self) -> float:
        return self.FREQ

    def intensity(self, temperature: float = -math.inf) -> float:
        if self.DR >= 0 and temperature > 0. and temperature != T0:
            return self.INT + (0.5 * self.DR + 1.0) * math.log10(T0 / temperature) + (
                    -(1 / temperature - 1 / T0) * self.ELO * 100. * h * c / k) * M_LOG10E
        else:
            return self.INT

    @property
    def degrees_of_freedom(self) -> int:
        return self.DR

    @property
    def lower_state_energy(self) -> float:
        return self.ELO

    def to_dict(self) -> dict[str, float]:
        return {FREQUENCY: self.FREQ, INTENSITY: self.INT, LOWER_STATE_ENERGY: self.ELO}

    def __repr__(self) -> str:
        return f'{self.FREQ} {self.INT} {self.ELO}'



# -*- coding: utf-8 -*-
from __future__ import annotations

from math import inf
from queue import Queue
from typing import cast

from qtpy.QtCore import QTimer, Slot
from qtpy.QtWidgets import (QDialog, QDoubleSpinBox, QFormLayout, QLabel, QProgressBar, QVBoxLayout,
                            QWidget, QWizard, QWizardPage)
from qtpy.compat import getsavefilename

try:
    from ..async_downloader import Downloader
except (SyntaxError, ImportError, ModuleNotFoundError):
    from ..downloader import Downloader
from .waiting_screen import WaitingScreen
from ..utils import all_cases, ensure_prefix, save_catalog_to_file


__all__ = ['DownloadDialog']


class SettingsPage(QWizardPage):
    def __init__(self, parent: QWidget | None = None) -> None:
        super().__init__(parent)

        self.setTitle(self.tr('New catalog'))

        layout: QFormLayout = QFormLayout(self)

        self.spin_min_frequency: QDoubleSpinBox = QDoubleSpinBox(self)
        self.spin_max_frequency: QDoubleSpinBox = QDoubleSpinBox(self)
        self.spin_min_frequency.setRange(0., inf)
        self.spin_max_frequency.setRange(0., inf)
        self.spin_min_frequency.valueChanged.connect(self.spin_max_frequency.setMinimum)
        self.spin_max_frequency.valueChanged.connect(self.spin_min_frequency.setMaximum)
        self.spin_min_frequency.setPrefix(self.tr('', 'spin prefix'))
        self.spin_max_frequency.setPrefix(self.tr('', 'spin prefix'))
        self.spin_min_frequency.setSuffix(self.tr(' MHz', 'spin suffix'))
        self.spin_max_frequency.setSuffix(self.tr(' MHz', 'spin suffix'))
        layout.addRow(self.tr('Minimal frequency:'), self.spin_min_frequency)
        layout.addRow(self.tr('Maximal frequency:'), self.spin_max_frequency)

        try:  # PyQt*
            self.registerField('min_frequency', self.spin_min_frequency,
                               'value', self.spin_min_frequency.valueChanged)
            self.registerField('max_frequency', self.spin_max_frequency,
                               'value', self.spin_max_frequency.valueChanged)
        except TypeError:  # PySide*
            self.registerField('min_frequency', self.spin_min_frequency,
                               'value', 'self.spin_min_frequency.valueChanged')
            self.registerField('max_frequency', self.spin_max_frequency,
                               'value', 'self.spin_max_frequency.valueChanged')

    @property
    def frequency_limits(self) -> tuple[float, float]:
        return self.spin_min_frequency.value(), self.spin_max_frequency.value()

    @frequency_limits.setter
    def frequency_limits(self, new_limits: tuple[float, float]) -> None:
        min_frequency: float = min(new_limits)
        max_frequency: float = max(new_limits)
        self.spin_max_frequency.setMaximum(max(2.0 * max_frequency, self.spin_max_frequency.maximum()))
        if min_frequency > self.spin_max_frequency.value():
            self.spin_max_frequency.setValue(max_frequency)
            self.spin_min_frequency.setValue(min_frequency)
        else:
            self.spin_min_frequency.setValue(min_frequency)
            self.spin_max_frequency.setValue(max_frequency)

    def validatePage(self) -> bool:
        return self.spin_max_frequency.value() > self.spin_min_frequency.value()


class DownloadConfirmationPage(QWizardPage):
    def __init__(self, parent: QWidget | None = None) -> None:
        super().__init__(parent)

        self.setTitle(self.tr('Downloading catalog'))
        self.setCommitPage(True)

        layout: QVBoxLayout = QVBoxLayout(self)
        self._label: QLabel = QLabel(self)
        layout.addWidget(self._label)

    def initializePage(self) -> None:
        super(DownloadConfirmationPage, self).initializePage()
        self.setButtonText(QWizard.WizardButton.CommitButton, self.tr('&Start'))
        self._label.setText(self.tr('Click {button_text} to start the download data'
                                    ' for {min_frequency} to {max_frequency} MHz.')
                            .format(button_text=self.buttonText(QWizard.WizardButton.CommitButton).replace('&', ''),
                                    min_frequency=self.field('min_frequency'),
                                    max_frequency=self.field('max_frequency')))


class ProgressPage(QWizardPage):
    def __init__(self, parent: QWidget | None = None) -> None:
        super().__init__(parent)

        self.setTitle(self.tr('Downloading catalog'))
        self.setCommitPage(True)

        layout: QVBoxLayout = QVBoxLayout(self)
        self.downloader: Downloader | None = None
        self.state_queue: Queue[tuple[int, int]] = Queue()

        self.progress_bar: QProgressBar = QProgressBar(self)
        layout.addWidget(self.progress_bar)

        self.timer: QTimer = QTimer(self)
        self.timer.timeout.connect(self._on_timeout)

    def initializePage(self) -> None:
        super(ProgressPage, self).initializePage()
        self.setButtonText(QWizard.WizardButton.CommitButton, self.buttonText(QWizard.WizardButton.NextButton))

        frequency_limits: tuple[float, float] = (
            self.field('min_frequency'),
            self.field('max_frequency'),
        )
        self.downloader = Downloader(frequency_limits=frequency_limits, state_queue=self.state_queue)
        self.downloader.start()
        self.timer.start(100)

    @Slot()
    def _on_timeout(self) -> None:
        while not self.state_queue.empty():
            cataloged_species: int
            not_yet_processed_species: int
            cataloged_species, not_yet_processed_species = self.state_queue.get(block=False)
            self.progress_bar.setValue(cataloged_species)
            self.progress_bar.setMaximum(cataloged_species + not_yet_processed_species)
        if self.isComplete():
            self.timer.stop()
            self.wizard().catalog = self.downloader.catalog
            self.completeChanged.emit()

    def isComplete(self) -> bool:
        return self.downloader is not None and not self.downloader.is_alive()


class SummaryPage(QWizardPage):
    def __init__(self, parent: QWidget | None = None) -> None:
        super().__init__(parent)
        layout: QVBoxLayout = QVBoxLayout(self)
        self._label: QLabel = QLabel(self)
        layout.addWidget(self._label)

    def initializePage(self) -> None:
        super(SummaryPage, self).initializePage()
        if cast(DownloadDialog, self.wizard()).catalog:
            self.setTitle(self.tr('Success'))
            self.setButtonText(QWizard.WizardButton.FinishButton, self.tr('&Save'))
            self._label.setText(
                self.tr('Click {button_text} to save the catalog'
                        ' for {min_frequency} to {max_frequency} MHz.')
                .format(button_text=self.buttonText(QWizard.WizardButton.FinishButton).replace('&', ''),
                        min_frequency=self.field('min_frequency'),
                        max_frequency=self.field('max_frequency')))
        else:
            self.setTitle(self.tr('Failure'))
            self._label.setText(self.tr('For the specified frequency range, nothing has been loaded.'))


class DownloadDialog(QWizard):
    """ GUI for `async_downloader.Downloader` """

    def __init__(self, frequency_limits: tuple[float, float] = (-inf, inf),
                 parent: QWidget | None = None) -> None:
        super().__init__(parent)

        self.catalog: list[dict[str, int | str | list[dict[str, float]]]] = []

        self.setModal(True)
        self.setWindowTitle(self.tr('Download Catalog'))
        if parent is not None:
            self.setWindowIcon(parent.windowIcon())

        self.settings_page: SettingsPage = SettingsPage(self)
        self.settings_page.frequency_limits = frequency_limits
        self.addPage(self.settings_page)
        self.addPage(DownloadConfirmationPage(self))
        self.progress_page: ProgressPage = ProgressPage(self)
        self.addPage(self.progress_page)
        self.addPage(SummaryPage(self))

    def back(self) -> None:
        if self.progress_page.downloader is not None and self.progress_page.downloader.is_alive():
            self.progress_page.downloader.join(0.1)
        super(DownloadDialog, self).back()

    def next(self) -> None:
        if self.progress_page.downloader is not None and self.progress_page.downloader.is_alive():
            self.progress_page.downloader.join(0.1)
        super(DownloadDialog, self).next()

    def restart(self) -> None:
        if self.progress_page.downloader is not None and self.progress_page.downloader.is_alive():
            self.progress_page.downloader.join(0.1)
        super(DownloadDialog, self).restart()

    def _get_save_file_name(self, formats: dict[tuple[str, ...], str],
                            caption: str = '', directory: str = '') -> tuple[str, str]:

        def join_file_dialog_formats(_formats: dict[tuple[str, ...], str]) -> str:
            f: tuple[str, ...]
            all_supported_extensions: list[str] = []
            for f in _formats.keys():
                all_supported_extensions.extend(ensure_prefix(_f, '*') for _f in f)
            format_lines: list[str] = [''.join((
                self.tr('All supported', 'file type'),
                '(',
                ' '.join(ensure_prefix(_f, '*') for _f in all_supported_extensions),
                ')'))]
            n: str
            for f, n in _formats.items():
                format_lines.append(''.join((n, '(', ' '.join(ensure_prefix(_f, '*') for _f in f), ')')))
            format_lines.append(self.tr('All files', 'file type') + '(* *.*)')
            return ';;'.join(format_lines)

        filename: str
        _filter: str
        filename, _filter = getsavefilename(self,
                                            caption=caption,
                                            filters=join_file_dialog_formats(formats),
                                            basedir=directory)
        return filename, _filter

    def done(self, exit_code: QDialog.DialogCode) -> None:
        if self.progress_page.downloader is not None and self.progress_page.downloader.is_alive():
            self.progress_page.downloader.join(0.1)

        if exit_code == QDialog.DialogCode.Accepted and self.catalog:
            _formats: dict[tuple[str, ...], str] = {
                tuple(all_cases('.json.gz')): self.tr('JSON with GZip compression', 'file type'),
                tuple(all_cases('.json.bz2')): self.tr('JSON with Bzip2 compression', 'file type'),
                (*all_cases('.json.xz'), *all_cases('.json.lzma')):
                    self.tr('JSON with LZMA2 compression', 'file type'),
                tuple(all_cases('.json')): self.tr('JSON', 'file type'),
            }
            save_file_name: str
            save_file_name, _ = self._get_save_file_name(formats=_formats, caption=self.tr('Save As...'))
            if not save_file_name:
                return

            ws: WaitingScreen = WaitingScreen(
                self,
                label=self.tr('Please wait...'),
                target=save_catalog_to_file,
                kwargs={
                    'filename': save_file_name,
                    'catalog': self.catalog,
                    'frequency_limits': (self.field('min_frequency'), self.field('max_frequency'))
                })
            ws.exec()

        super(DownloadDialog, self).done(exit_code)



# -*- coding: utf-8 -*-
from __future__ import annotations

import html
import html.entities
import itertools
import math
import os
import sys
from numbers import Real
from pathlib import Path
from typing import Any, Callable, Final, Iterable, Iterator, Protocol, Sequence, TypeVar, cast, overload

__all__ = ['M_LOG10E',
           'T0', 'c', 'h', 'k', 'e',
           'CATALOG', 'BUILD_TIME', 'LINES', 'FREQUENCY', 'INTENSITY', 'ID', 'STRUCTURAL_FORMULA',
           'STOICHIOMETRIC_FORMULA', 'MOLECULE', 'MOLECULE_SYMBOL', 'SPECIES_TAG', 'NAME', 'TRIVIAL_NAME', 'ISOTOPOLOG',
           'STATE', 'STATE_HTML', 'INCHI_KEY', 'DEGREES_OF_FREEDOM', 'LOWER_STATE_ENERGY', 'CONTRIBUTOR', 'VERSION',
           'DATE_OF_ENTRY', 'HUMAN_READABLE',
           'ghz_to_mhz', 'ghz_to_nm', 'ghz_to_rec_cm',
           'mhz_to_ghz', 'mhz_to_nm', 'mhz_to_rec_cm',
           'nm_to_ghz', 'nm_to_mhz', 'nm_to_rec_cm',
           'rec_cm_to_ghz', 'rec_cm_to_mhz', 'rec_cm_to_nm', 'rec_cm_to_meV', 'rec_cm_to_j',
           'meV_to_rec_cm', 'j_to_rec_cm',
           'log10_sq_nm_mhz_to_sq_nm_mhz',
           'log10_sq_nm_mhz_to_log10_cm_per_molecule',
           'log10_sq_nm_mhz_to_cm_per_molecule',
           'sq_nm_mhz_to_log10_sq_nm_mhz',
           'log10_cm_per_molecule_to_log10_sq_nm_mhz',
           'cm_per_molecule_to_log10_sq_nm_mhz',
           'sort_unique', 'merge_sorted', 'search_sorted',
           'within', 'chem_html', 'best_name', 'remove_html', 'wrap_in_html',
           'ensure_prefix', 'all_cases',
           'save_catalog_to_file',
           'ReleaseInfo', 'latest_release', 'update_with_pip']

M_LOG10E: Final[float] = math.log10(math.e)

T0: Final[float] = 300.00  # [K], see https://spec.jpl.nasa.gov/ftp/pub/catalog/doc/catdoc.pdf
k: Final[float] = 1.380649000e-23  # [J/K],  see https://physics.nist.gov/cgi-bin/cuu/Value?k
h: Final[float] = 6.626070150e-34  # [J/Hz], see https://physics.nist.gov/cgi-bin/cuu/Value?h
e: Final[float] = 1.602176634e-19  # [C],    see https://physics.nist.gov/cgi-bin/cuu/Value?e
c: Final[float] = 299_792_458.000  # [m/s],  see https://physics.nist.gov/cgi-bin/cuu/Value?c

CATALOG: Final[str] = 'catalog'
BUILD_TIME: Final[str] = 'build_time'
LINES: Final[str] = 'lines'
FREQUENCY: Final[str] = 'frequency'
INTENSITY: Final[str] = 'intensity'
ID: Final[str] = 'id'
MOLECULE: Final[str] = 'molecule'
STRUCTURAL_FORMULA: Final[str] = 'structuralformula'
STOICHIOMETRIC_FORMULA: Final[str] = 'stoichiometricformula'
MOLECULE_SYMBOL: Final[str] = 'moleculesymbol'
SPECIES_TAG: Final[str] = 'speciestag'
NAME: Final[str] = 'name'
TRIVIAL_NAME: Final[str] = 'trivialname'
ISOTOPOLOG: Final[str] = 'isotopolog'
STATE: Final[str] = 'state'
STATE_HTML: Final[str] = 'state_html'
INCHI_KEY: Final[str] = 'inchikey'
CONTRIBUTOR: Final[str] = 'contributor'
VERSION: Final[str] = 'version'
DATE_OF_ENTRY: Final[str] = 'dateofentry'
DEGREES_OF_FREEDOM: Final[str] = 'degreesoffreedom'
LOWER_STATE_ENERGY: Final[str] = 'lowerstateenergy'

HUMAN_READABLE: Final[dict[str, str]] = {
    CATALOG: 'Catalog',
    LINES: 'Lines',
    FREQUENCY: 'Frequency',
    INTENSITY: 'Intensity',
    ID: 'ID',
    MOLECULE: 'Molecule',
    STRUCTURAL_FORMULA: 'Structural formula',
    STOICHIOMETRIC_FORMULA: 'Stoichiometric formula',
    MOLECULE_SYMBOL: 'Molecule symbol',
    SPECIES_TAG: 'Species tag',
    NAME: 'Name',
    TRIVIAL_NAME: 'Trivial name',
    ISOTOPOLOG: 'Isotopolog',
    STATE: 'State (TeX)',
    STATE_HTML: 'State (HTML)',
    INCHI_KEY: 'InChI key',
    CONTRIBUTOR: 'Contributor',
    VERSION: 'Version',
    DATE_OF_ENTRY: 'Date of entry',
    DEGREES_OF_FREEDOM: 'Degrees of freedom',
    LOWER_STATE_ENERGY: 'Lower state energy',
}


def within(x: float, limits: tuple[float, float] | tuple[tuple[float, float], ...]) -> bool:
    if len(limits) < 2:
        raise ValueError('Invalid limits')
    if all(isinstance(limit, Real) for limit in limits):
        return min(limits) <= x <= max(limits)
    elif all(isinstance(limit, tuple) for limit in limits):
        return any(min(limit) <= x <= max(limit) for limit in limits)
    else:
        raise TypeError('Invalid limits type')


_AnyType = TypeVar('_AnyType')


class SupportsLessAndEqual(Protocol[_AnyType]):
    def __eq__(self, other: _AnyType) -> bool:
        pass

    def __lt__(self, other: _AnyType) -> bool:
        pass

    def __le__(self, other: _AnyType) -> bool:
        pass


@overload
def sort_unique(items: Sequence[_AnyType], *,
                key: Callable[[_AnyType], SupportsLessAndEqual],
                reverse: bool = False) -> list[_AnyType]:
    pass


@overload
def sort_unique(items: Sequence[SupportsLessAndEqual], *,
                key: Callable[[SupportsLessAndEqual], SupportsLessAndEqual] | None = None,
                reverse: bool = False) -> list[SupportsLessAndEqual]:
    pass


def sort_unique(items: Sequence[SupportsLessAndEqual] | Sequence[_AnyType], *,
                key: Callable[[_AnyType], SupportsLessAndEqual] | None = None,
                reverse: bool = False) -> list[SupportsLessAndEqual]:
    sorted_items: list[SupportsLessAndEqual] = sorted(items, key=key, reverse=reverse)
    i: int = 0
    while i < len(sorted_items) - 1:
        while i < len(sorted_items) - 1 and sorted_items[i] == sorted_items[i + 1]:
            del sorted_items[i + 1]
        i += 1
    return sorted_items


@overload
def merge_sorted(items_1: Sequence[_AnyType],
                 items_2: Sequence[_AnyType], *,
                 key: Callable[[_AnyType], SupportsLessAndEqual]) -> list[_AnyType]:
    pass


@overload
def merge_sorted(items_1: Sequence[SupportsLessAndEqual],
                 items_2: Sequence[SupportsLessAndEqual], *,
                 key: Callable[[_AnyType], SupportsLessAndEqual] | None = None) -> list[SupportsLessAndEqual]:
    pass


def merge_sorted(items_1: Sequence[SupportsLessAndEqual] | Sequence[_AnyType],
                 items_2: Sequence[SupportsLessAndEqual] | Sequence[_AnyType], *,
                 key: Callable[[_AnyType], SupportsLessAndEqual] | None = None) -> list[SupportsLessAndEqual]:
    sorted_items_1: list[SupportsLessAndEqual] = sort_unique(items_1, key=key, reverse=False)
    sorted_items_2: list[SupportsLessAndEqual] = sort_unique(items_2, key=key, reverse=False)
    merged_items: list[SupportsLessAndEqual] = []

    last_i_1: int
    last_i_2: int
    i_1: int = 0
    i_2: int = 0

    if key is None:
        def key(value: _AnyType) -> SupportsLessAndEqual:
            return value

    while i_1 < len(sorted_items_1) and i_2 < len(sorted_items_2):
        last_i_1 = i_1
        while (i_1 < len(sorted_items_1) and i_2 < len(sorted_items_2)
               and key(sorted_items_1[i_1]) <= key(sorted_items_2[i_2])):
            i_1 += 1
        if last_i_1 < i_1:
            merged_items.extend(sorted_items_1[last_i_1:i_1])

        last_i_2 = i_2
        while (i_1 < len(sorted_items_1) and i_2 < len(sorted_items_2)
               and key(sorted_items_2[i_2]) <= key(sorted_items_1[i_1])):
            i_2 += 1
        if last_i_2 < i_2:
            merged_items.extend(sorted_items_2[last_i_2:i_2])

    while i_1 < len(sorted_items_1) and key(merged_items[-1]) == key(sorted_items_1[i_1]):
        i_1 += 1
    if i_1 < len(sorted_items_1) and i_2 >= len(sorted_items_2):
        merged_items.extend(sorted_items_1[i_1:])
    while i_2 < len(sorted_items_2) and key(merged_items[-1]) == key(sorted_items_2[i_2]):
        i_2 += 1
    if i_2 < len(sorted_items_2) and i_1 >= len(sorted_items_1):
        merged_items.extend(sorted_items_2[i_2:])
    return merged_items


@overload
def search_sorted(threshold: SupportsLessAndEqual,
                  items: Sequence[_AnyType], *,
                  key: Callable[[_AnyType], SupportsLessAndEqual],
                  maybe_equal: bool = False) -> int:
    pass


@overload
def search_sorted(threshold: SupportsLessAndEqual,
                  items: Sequence[SupportsLessAndEqual], *,
                  key: Callable[[_AnyType], SupportsLessAndEqual] | None = None,
                  maybe_equal: bool = False) -> int:
    pass


def search_sorted(threshold: SupportsLessAndEqual,
                  items: Sequence[_AnyType] | Sequence[SupportsLessAndEqual], *,
                  key: Callable[[_AnyType], SupportsLessAndEqual] | None = None,
                  maybe_equal: bool = False) -> int:
    from operator import lt, le

    if not items:
        raise ValueError('Empty sequence provided')
    if key is None:
        def key(value: _AnyType) -> SupportsLessAndEqual:
            return value

    less: Callable[[SupportsLessAndEqual, SupportsLessAndEqual], bool] = le if maybe_equal else lt
    if not less(key(items[0]), threshold):
        return -1
    if less(key(items[-1]), threshold):
        return len(items)
    i: int = 0
    j: int = len(items) - 1
    n: int
    while j - i > 1:
        n = (i + j) // 2
        if less(key(items[n]), threshold):
            i = n
        else:
            j = n
    if i != j and not less(key(items[j]), threshold):
        return i
    return j


def mhz_to_ghz(frequency_mhz: float) -> float:
    return frequency_mhz * 1e-3


def mhz_to_rec_cm(frequency_mhz: float) -> float:
    return frequency_mhz * 1e4 / c


def mhz_to_nm(frequency_mhz: float) -> float:
    return c / frequency_mhz * 1e3


def ghz_to_mhz(frequency_ghz: float) -> float:
    return frequency_ghz * 1e3


def ghz_to_rec_cm(frequency_ghz: float) -> float:
    return frequency_ghz * 1e7 / c


def ghz_to_nm(frequency_ghz: float) -> float:
    return c / frequency_ghz


def rec_cm_to_mhz(frequency_rec_cm: float) -> float:
    return frequency_rec_cm * 1e-4 * c


def rec_cm_to_ghz(frequency_rec_cm: float) -> float:
    return frequency_rec_cm * 1e-7 * c


def rec_cm_to_nm(frequency_rec_cm: float) -> float:
    return 1e7 / frequency_rec_cm


def rec_cm_to_meV(energy_rec_cm: float) -> float:
    return 1e5 * h * c / e * energy_rec_cm


def rec_cm_to_j(energy_rec_cm: float) -> float:
    return 1e2 * h * c * energy_rec_cm


def nm_to_mhz(frequency_nm: float) -> float:
    return c / frequency_nm * 1e-3


def nm_to_ghz(frequency_nm: float) -> float:
    return c / frequency_nm


def nm_to_rec_cm(frequency_nm: float) -> float:
    return 1e7 / frequency_nm


def meV_to_rec_cm(energy_mev: float) -> float:
    return 1e-5 * e / h / c * energy_mev


def j_to_rec_cm(energy_j: float) -> float:
    return 1e-2 / h / c * energy_j


def log10_sq_nm_mhz_to_sq_nm_mhz(intensity_log10_sq_nm_mhz: float) -> float:
    return math.pow(10.0, intensity_log10_sq_nm_mhz)


def log10_sq_nm_mhz_to_log10_cm_per_molecule(intensity_log10_sq_nm_mhz: float) -> float:
    return -10. + intensity_log10_sq_nm_mhz - math.log10(c)


def log10_sq_nm_mhz_to_cm_per_molecule(intensity_log10_sq_nm_mhz: float) -> float:
    return math.pow(10.0, log10_sq_nm_mhz_to_log10_cm_per_molecule(intensity_log10_sq_nm_mhz))


def sq_nm_mhz_to_log10_sq_nm_mhz(intensity_sq_nm_mhz: float) -> float:
    if intensity_sq_nm_mhz == 0.0:
        return -math.inf
    if intensity_sq_nm_mhz < 0.0:
        return math.nan
    return math.log10(intensity_sq_nm_mhz)


def log10_cm_per_molecule_to_log10_sq_nm_mhz(intensity_log10_cm_per_molecule: float) -> float:
    return intensity_log10_cm_per_molecule + 10. + math.log10(c)


def cm_per_molecule_to_log10_sq_nm_mhz(intensity_cm_per_molecule: float) -> float:
    if intensity_cm_per_molecule == 0.0:
        return -math.inf
    if intensity_cm_per_molecule < 0.0:
        return math.nan
    return log10_cm_per_molecule_to_log10_sq_nm_mhz(math.log10(intensity_cm_per_molecule))


def tex_to_html_entity(s: str) -> str:
    r"""
    Change LaTeX entities syntax to HTML one.
    Get ‘\alpha’ to be ‘&alpha;’ and so on. Unknown LaTeX entities do not get replaced.

    :param s: a line to convert
    :return: a line with all LaTeX entities renamed
    """
    word_start: int = -1
    word_started: bool = False
    backslash_found: bool = False
    _i: int = 0
    fixes: dict[str, str] = {
        'neq': '#8800',
    }
    while _i < len(s):
        _c: str = s[_i]
        if word_started and not _c.isalpha():
            word_started = False
            if s[word_start:_i] + ';' in html.entities.entitydefs:
                s = s[:word_start - 1] + '&' + s[word_start:_i] + ';' + s[_i:]
                _i += 2
            elif s[word_start:_i] in fixes:
                s = s[:word_start - 1] + '&' + fixes[s[word_start:_i]] + ';' + s[_i:]
                _i += 2
        if backslash_found and _c.isalpha() and not word_started:
            word_start = _i
            word_started = True
        backslash_found = (_c == '\\')
        _i += 1
    if word_started:
        if s[word_start:_i] + ';' in html.entities.entitydefs:
            s = s[:word_start - 1] + '&' + s[word_start:_i] + ';' + s[_i:]
            _i += 2
        elif s[word_start:_i] in fixes:
            s = s[:word_start - 1] + '&' + fixes[s[word_start:_i]] + ';' + s[_i:]
            _i += 2
    return s


def chem_html(formula: str) -> str:
    """ converts plain text chemical formula into html markup """
    if '<' in formula or '>' in formula:
        # we can not tell whether it's a tag or a mathematical sign
        return formula

    def subscript(s: str) -> str:
        number_start: int = -1
        number_started: bool = False
        cap_alpha_started: bool = False
        low_alpha_started: bool = False
        _i: int = 0
        while _i < len(s):
            _c: str = s[_i]
            if number_started and not _c.isdigit():
                number_started = False
                s = s[:number_start] + '<sub>' + s[number_start:_i] + '</sub>' + s[_i:]
                _i += 11
            if (cap_alpha_started or low_alpha_started) and _c.isdigit() and not number_started:
                number_start = _i
                number_started = True
            if low_alpha_started:
                cap_alpha_started = False
                low_alpha_started = False
            if cap_alpha_started and _c.islower() or _c == ')':
                low_alpha_started = True
            cap_alpha_started = _c.isupper()
            _i += 1
        if number_started:
            s = s[:number_start] + '<sub>' + s[number_start:] + '</sub>'
        return s

    def prefix(s: str) -> str:
        no_digits: bool = False
        _i: int = len(s)
        while not no_digits:
            _i = s.rfind('-', 0, _i)
            if _i == -1:
                break
            if s[:_i].isalpha() and s[:_i].isupper():
                break
            no_digits = True
            _c: str
            unescaped_prefix: str = html.unescape(s[:_i])
            for _c in unescaped_prefix:
                if _c.isdigit() or _c == '<':
                    no_digits = False
                    break
            if no_digits and (unescaped_prefix[0].islower() or unescaped_prefix[0] == '('):
                return '<i>' + s[:_i] + '</i>' + s[_i:]
        return s

    def charge(s: str) -> str:
        if s[-1] in '+-':
            return s[:-1] + '<sup>' + s[-1] + '</sup>'
        return s

    def v(s: str) -> str:
        if '=' not in s:
            return s[0] + ' = ' + s[1:]
        ss: list[str] = list(map(str.strip, s.split('=')))
        for _i in range(len(ss)):
            if ss[_i].startswith('v'):
                ss[_i] = ss[_i][0] + '<sub>' + ss[_i][1:] + '</sub>'
        return ' = '.join(ss)

    html_formula: str = html.escape(formula)
    html_formula_pieces: list[str] = list(map(str.strip, html_formula.split(',')))
    for i in range(len(html_formula_pieces)):
        if html_formula_pieces[i].startswith('v'):
            html_formula_pieces = html_formula_pieces[:i] + [', '.join(html_formula_pieces[i:])]
            break
    for i in range(len(html_formula_pieces)):
        if html_formula_pieces[i].startswith('v'):
            html_formula_pieces[i] = v(html_formula_pieces[i])
            break
        for function in (subscript, prefix, charge):
            html_formula_pieces[i] = function(html_formula_pieces[i])
    html_formula = ', '.join(html_formula_pieces)
    return html_formula


def is_good_html(text: str) -> bool:
    """ Basic check that all tags are sound """
    _1, _2, _3 = text.count('<'), text.count('>'), 2 * text.count('</')
    return _1 == _2 and _1 == _3


def best_name(entry: dict[str, int | str | list[dict[str, float]]],
              allow_html: bool = True) -> str:
    species_tag: int = cast(int, entry.get(SPECIES_TAG, 0))
    last: str = best_name.__dict__.get('last', dict()).get(species_tag, dict()).get(allow_html, '')
    if last:
        return last

    def _best_name() -> str:
        if isotopolog := entry.get(ISOTOPOLOG, ''):
            if allow_html:
                if (is_good_html(str(molecule_symbol := entry[MOLECULE_SYMBOL]))
                        and (entry.get(STRUCTURAL_FORMULA, '') == isotopolog
                             or entry.get(STOICHIOMETRIC_FORMULA, '') == isotopolog)):
                    if state_html := cast(str, entry.get(STATE_HTML, '')):
                        # span tags are needed when the molecule symbol is malformed
                        return f'<span>{molecule_symbol}</span>, ' \
                               f'{chem_html(tex_to_html_entity(str(state_html)))}'
                    return str(molecule_symbol)
                else:
                    if state_html := cast(str, entry.get(STATE_HTML, '')):
                        return f'{chem_html(str(isotopolog))}, ' \
                               f'{chem_html(tex_to_html_entity(str(state_html)))}'
                    return chem_html(str(isotopolog))
            else:
                if state_html := cast(str, entry.get(STATE_HTML, '')):
                    return f'{isotopolog}, {remove_html(tex_to_html_entity(state_html))}'
                if state := cast(str, entry.get(STATE, '')):
                    return f'{isotopolog}, {remove_html(tex_to_html_entity(state.strip("$")))}'
                return isotopolog

        for key in (NAME, STRUCTURAL_FORMULA, STOICHIOMETRIC_FORMULA):
            if candidate := entry.get(key, ''):
                return chem_html(str(candidate)) if allow_html else str(candidate)
        if trivial_name := entry.get(TRIVIAL_NAME, ''):
            return str(trivial_name)
        if species_tag:
            return str(species_tag)
        return 'no name'

    res: str = _best_name()
    if not species_tag:
        return res
    if 'last' not in best_name.__dict__:
        best_name.__dict__['last'] = dict()
    if species_tag not in best_name.__dict__['last']:
        best_name.__dict__['last'][species_tag] = dict()
    best_name.__dict__['last'][species_tag][allow_html] = res
    return res


def remove_html(line: str) -> str:
    """ removes HTML tags and decodes HTML entities """
    if not is_good_html(line):
        return html.unescape(line)

    new_line: str = line
    tag_start: int = new_line.find('<')
    tag_end: int = new_line.find('>', tag_start)
    while tag_start != -1 and tag_end != -1:
        new_line = new_line[:tag_start] + new_line[tag_end + 1:]
        tag_start = new_line.find('<')
        tag_end = new_line.find('>', tag_start)
    return html.unescape(new_line).lstrip()


def wrap_in_html(text: str, line_end: str = os.linesep) -> str:
    """ Make a full HTML document out of a piece of the markup """
    new_text: list[str] = [
        '<!DOCTYPE HTML PUBLIC "-//W3C//DTD HTML 4.0 Transitional//EN">',
        '<html lang="en" xml:lang="en">',
        '<head>',
        '<meta http-equiv="content-type" content="text/html; charset=utf-8">',
        '</head>',
        '<body>',
        text,
        '</body>',
        '</html>'
    ]

    return line_end.join(new_text)


def ensure_prefix(text: str, prefix: str) -> str:
    if text.startswith(prefix):
        return text
    else:
        return prefix + text


def all_cases(text: str) -> Iterator[str]:
    """ return all cases of the text given """

    cases: list[str] = list({text.lower(), text.upper(), text.capitalize(), text.swapcase(), text.casefold()})

    if len(cases) < 2:
        # don't bother
        yield from cases
        return

    length: int = len(cases[0])
    case: str
    if not all(len(case) == length for case in cases):
        # don't know what to do with cases of different lengths
        yield from cases
        return
    # now, all the cases are of the same length

    # get all possible variants of characters at each position
    i: int
    variants: Iterator[list[str]] = (sorted(set(case[i] for case in cases), reverse=True) for i in range(length))

    combination: tuple[str, ...]
    for combination in itertools.product(*variants):
        yield ''.join(combination)


def save_catalog_to_file(filename: str | Path,
                         catalog: list[dict[str, int | str | list[dict[str, float]]]],
                         frequency_limits: tuple[float, float]) -> bool:
    from .catalog import Catalog

    if not catalog:
        return False
    Catalog.save(filename=filename, catalog=catalog, frequency_limits=frequency_limits)
    return True


class ReleaseInfo:
    def __init__(self, version: str = '', pub_date: str = '') -> None:
        self.version: str = version
        self.pub_date: str = pub_date

    def __bool__(self) -> bool:
        return bool(self.version) and bool(self.pub_date)

    def __lt__(self, other: str | ReleaseInfo) -> bool:
        if isinstance(other, str):
            other = ReleaseInfo(version=other)
        i: str
        j: str
        for i, j in itertools.zip_longest(self.version.replace('-', '.').split('.'),
                                          other.version.replace('-', '.').split('.'), fillvalue=''):
            if i == j:
                continue
            if i.isdigit() and j.isdigit():
                return int(i) < int(j)
            else:
                i_digits: str = ''.join(itertools.takewhile(str.isdigit, i))
                j_digits: str = ''.join(itertools.takewhile(str.isdigit, j))
                if i_digits != j_digits:
                    if i_digits and j_digits:
                        return int(i_digits) < int(j_digits)
                    else:
                        return i_digits < j_digits
                return i < j
        return False

    def __eq__(self, other: str | ReleaseInfo) -> bool:
        if not isinstance(other, (str, ReleaseInfo)):
            return NotImplemented
        if isinstance(other, str):
            other = ReleaseInfo(version=other)
        return self.version == other.version


def latest_release() -> ReleaseInfo:
    import urllib.request
    import xml.dom.minidom as dom
    from http.client import HTTPResponse
    from xml.dom.minicompat import NodeList
    from urllib.error import URLError

    from . import __original_name__

    try:
        r: HTTPResponse = urllib.request.urlopen(f'https://pypi.org/rss/project/{__original_name__}/releases.xml',
                                                 timeout=1)
    except URLError:
        return ReleaseInfo()
    if r.getcode() != 200 or not r.readable():
        return ReleaseInfo()
    rss: dom.Node | None = dom.parseString(r.read().decode(encoding='ascii')).documentElement
    if not isinstance(rss, dom.Element) or rss.tagName != 'rss':
        return ReleaseInfo()
    channels: NodeList = rss.getElementsByTagName('channel')
    if not channels or channels[0].nodeType != dom.Node.ELEMENT_NODE:
        return ReleaseInfo()
    channel: dom.Element = channels[0]
    items: NodeList = channel.getElementsByTagName('item')
    if not items or items[0].nodeType != dom.Node.ELEMENT_NODE:
        return ReleaseInfo()
    item: dom.Element = items[0]
    titles: NodeList = item.getElementsByTagName('title')
    if not titles or titles[0].nodeType != dom.Node.ELEMENT_NODE:
        return ReleaseInfo()
    title: dom.Element = titles[0]
    pub_dates: NodeList = item.getElementsByTagName('pubDate')
    if not pub_dates or pub_dates[0].nodeType != dom.Node.ELEMENT_NODE:
        return ReleaseInfo()
    pub_date: dom.Element = pub_dates[0]
    title_value: dom.Node = title.firstChild
    pub_date_value: dom.Node = pub_date.firstChild
    if not isinstance(title_value, dom.Text) or not isinstance(pub_date_value, dom.Text):
        return ReleaseInfo()

    return ReleaseInfo(title_value.data, pub_date_value.data)


def update_with_pip() -> None:
    import subprocess
    import sys

    from . import __original_name__

    subprocess.Popen(args=[
        sys.executable, '-c',
        f'''import sys, subprocess, time; time.sleep(2);\
        subprocess.run(args=[sys.executable, '-m', 'pip', 'install', '-U', {__original_name__!r}]);\
        subprocess.Popen(args=[sys.executable, '-m', {__original_name__!r}])'''])
    sys.exit(0)


if sys.version_info < (3, 10):
    import builtins


    # noinspection PyShadowingBuiltins, PyUnusedLocal
    def zip(*iterables: Iterable[Any], strict: bool = False) -> builtins.zip:
        """ Intentionally override `builtins.zip` to ignore `strict` parameter in Python < 3.10 """
        return builtins.zip(*iterables)


    __all__.append('zip')



# -*- coding: utf-8 -*-
from __future__ import annotations

from typing import Iterable

from qtpy.QtCore import QAbstractTableModel, QDateTime, QModelIndex, QObject, QPersistentModelIndex, QTimeZone, Qt
from qtpy.QtWidgets import (QAbstractItemView, QDialog, QDialogButtonBox, QFormLayout, QHeaderView, QTableView,
                            QVBoxLayout, QWidget)

from .selectable_label import SelectableLabel
from .titled_list_widget import TitledListWidget
from ..catalog import Catalog, CatalogSourceInfo

__all__ = ['CatalogInfo']


class DataModel(QAbstractTableModel):
    def __init__(self, parent: QObject | None = None) -> None:
        super().__init__(parent)
        self._data: list[CatalogSourceInfo] = []
        self._header: list[str] = []

    @property
    def header(self) -> list[str]:
        return self._header[:self.columnCount()]

    @header.setter
    def header(self, new_header: Iterable[str]) -> None:
        self._header = list(new_header)

    def rowCount(self, parent: QModelIndex | QPersistentModelIndex = QModelIndex()) -> int:
        return len(self._data)

    def columnCount(self, parent: QModelIndex | QPersistentModelIndex = QModelIndex()) -> int:
        return 1 if all(info.build_datetime is None for info in self._data) else 2

    def data(self, index: QModelIndex | QPersistentModelIndex, role: int = Qt.ItemDataRole.DisplayRole) -> str | None:
        if index.isValid() and role == Qt.ItemDataRole.DisplayRole:
            if index.column() == 0:
                return self._data[index.row()].filename
            if index.column() == 1:
                info: CatalogSourceInfo = self._data[index.row()]
                if info.build_datetime is None:
                    return None
                qt_datetime: QDateTime = QDateTime(info.build_datetime)
                qt_datetime.setTimeZone(QTimeZone(round(info.build_datetime.utcoffset().total_seconds())))
                return qt_datetime.toString()
        return None

    def headerData(self, col: int, orientation: Qt.Orientation, role: int = Qt.ItemDataRole.DisplayRole) -> str | None:
        if (orientation == Qt.Orientation.Horizontal
                and role == Qt.ItemDataRole.DisplayRole and 0 <= col < len(self._header)):
            return str(self._header[col])
        return None

    def setHeaderData(self, section: int, orientation: Qt.Orientation, value: str,
                      role: int = Qt.ItemDataRole.DisplayRole) -> bool:
        if (orientation == Qt.Orientation.Horizontal
                and role == Qt.ItemDataRole.DisplayRole and 0 <= section < len(self._header)):
            self._header[section] = value
            return True
        return False

    def clear(self) -> None:
        self.beginResetModel()
        self._data = []
        self.endResetModel()

    def sort(self, column: int, order: Qt.SortOrder = Qt.SortOrder.AscendingOrder) -> None:
        self.beginResetModel()
        if column == 0:
            self._data.sort(key=lambda i: i.filename)
        elif column == 1:
            self._data.sort(key=lambda i: i.build_datetime)
        self.endResetModel()

    def append(self, info: CatalogSourceInfo) -> None:
        self.beginResetModel()
        self._data.append(info)
        self.endResetModel()

    def extend(self, info: Iterable[CatalogSourceInfo]) -> None:
        self.beginResetModel()
        self._data.extend(info)
        self.endResetModel()


class SourcesList(QTableView):
    def __init__(self, parent: QWidget | None = None) -> None:
        super().__init__(parent)
        self.setAlternatingRowColors(True)
        self._model: DataModel = DataModel()
        self._model.header = [self.tr('Filename'), self.tr('Build Time')]
        self.setModel(self._model)
        self.setEditTriggers(QAbstractItemView.EditTrigger.NoEditTriggers)
        self.setDropIndicatorShown(False)
        self.setDragDropOverwriteMode(False)
        self.setSelectionBehavior(QAbstractItemView.SelectionBehavior.SelectRows)
        self.setCornerButtonEnabled(False)
        self.setSortingEnabled(True)
        self.setAlternatingRowColors(True)
        self.horizontalHeader().setDefaultSectionSize(180)
        self.horizontalHeader().setHighlightSections(False)
        self.verticalHeader().setVisible(False)
        self.verticalHeader().setHighlightSections(False)

    def append(self, info: CatalogSourceInfo) -> None:
        self._model.append(info)
        if self.horizontalHeader().count() > 0:
            self.horizontalHeader().setSectionResizeMode(0, QHeaderView.ResizeMode.Stretch)
        for i in range(1, self.horizontalHeader().count()):
            self.horizontalHeader().setSectionResizeMode(i, QHeaderView.ResizeMode.ResizeToContents)
        self.resizeColumnsToContents()

    def extend(self, info: Iterable[CatalogSourceInfo]) -> None:
        self._model.extend(info)
        if self.horizontalHeader().count() > 0:
            self.horizontalHeader().setSectionResizeMode(0, QHeaderView.ResizeMode.Stretch)
        for i in range(1, self.horizontalHeader().count()):
            self.horizontalHeader().setSectionResizeMode(i, QHeaderView.ResizeMode.ResizeToContents)
        self.resizeColumnsToContents()


class CatalogInfo(QDialog):
    """ A simple dialog that displays the information about the loaded catalog(s) """

    def __init__(self, catalog: Catalog, parent: QWidget | None = None) -> None:
        super().__init__(parent)
        self.setModal(True)
        self.setWindowTitle(self.tr('Catalog Info'))
        if parent is not None:
            self.setWindowIcon(parent.windowIcon())
        layout: QVBoxLayout = QVBoxLayout(self)

        sources_list: SourcesList = SourcesList(self)
        layout.addWidget(sources_list)
        sources_list.extend(catalog.sources_info)

        frequency_limits_list: TitledListWidget = TitledListWidget(self)
        frequency_limits_list.setTitle(self.tr('Frequency limits:'))
        layout.addWidget(frequency_limits_list)
        frequency_limits_list.addItems([
            self.tr('{frequency_limit[0]} to {frequency_limit[1]} MHz').format(frequency_limit=frequency_limit)
            for frequency_limit in catalog.frequency_limits])

        stat_layout: QFormLayout = QFormLayout()
        stat_layout.addRow(self.tr('Total number of substances:'), SelectableLabel(str(catalog.entries_count)))
        layout.addLayout(stat_layout, 0)

        buttons: QDialogButtonBox = QDialogButtonBox(QDialogButtonBox.StandardButton.Close, self)
        buttons.rejected.connect(self.reject)
        layout.addWidget(buttons)



# -*- coding: utf-8 -*-
from __future__ import annotations

from math import inf

from qtpy.QtCore import Qt, Slot
from qtpy.QtWidgets import QAbstractSpinBox, QDoubleSpinBox, QFormLayout, QTabWidget, QWidget

from .settings import Settings
from ..utils import *

__all__ = ['FrequencyBox']


class FrequencySpinBox(QDoubleSpinBox):
    """`QDoubleSpinBox` with altered defaults"""
    def __init__(self, parent: QWidget | None = None) -> None:
        super().__init__(parent)
        self.setAlignment(Qt.AlignmentFlag.AlignRight
                          | Qt.AlignmentFlag.AlignTrailing
                          | Qt.AlignmentFlag.AlignVCenter)
        self.setButtonSymbols(QAbstractSpinBox.ButtonSymbols.NoButtons)
        self.setAccelerated(True)
        self.setDecimals(4)
        self.setMaximum(9999999.9999)
        self.setSuffix(self.tr(' MHz'))
        self.setCorrectionMode(QAbstractSpinBox.CorrectionMode.CorrectToNearestValue)


class FrequencyBox(QTabWidget):
    def __init__(self, settings: Settings, parent: QWidget | None = None) -> None:
        from . import icon  # import locally to avoid a circular import

        super().__init__(parent)

        self._settings: Settings = settings

        self._frequency_from: float = -inf  # [MHz]
        self._frequency_to: float = inf  # [MHz]
        self._frequency_center: float = 0.0  # [MHz]
        self._frequency_deviation: float = inf  # [MHz]

        self._page_by_range: QWidget = QWidget()
        self._layout_by_range: QFormLayout = QFormLayout(self._page_by_range)
        self._spin_frequency_from: FrequencySpinBox = FrequencySpinBox(self._page_by_range)
        self._spin_frequency_to: FrequencySpinBox = FrequencySpinBox(self._page_by_range)
        self._page_by_center: QWidget = QWidget()
        self._layout_by_center: QFormLayout = QFormLayout(self._page_by_center)
        self._spin_frequency_center: FrequencySpinBox = FrequencySpinBox(self._page_by_center)
        self._spin_frequency_deviation: FrequencySpinBox = FrequencySpinBox(self._page_by_center)

        self._layout_by_range.setLabelAlignment(Qt.AlignmentFlag.AlignLeft)
        self._spin_frequency_from.setValue(118747.341)
        self._layout_by_range.addRow(self._layout_by_range.tr('From:'), self._spin_frequency_from)
        self._spin_frequency_to.setValue(118753.341)
        self._layout_by_range.addRow(self._layout_by_range.tr('To:'), self._spin_frequency_to)
        self.addTab(self._page_by_range, icon('mdi6.arrow-expand-horizontal'), self.tr('Range'))

        self._spin_frequency_center.setValue(118750.341)
        self._layout_by_center.addRow(self._layout_by_center.tr('Center:'), self._spin_frequency_center)
        self._spin_frequency_deviation.setMaximum(99.9999)
        self._spin_frequency_deviation.setSingleStep(0.1)
        self._spin_frequency_deviation.setValue(0.4)
        self._layout_by_center.addRow(self._layout_by_center.tr('Deviation:'), self._spin_frequency_deviation)
        self.addTab(self._page_by_center, icon('mdi6.format-horizontal-align-center'), self.tr('Center'))

        self.load_settings()

        self._spin_frequency_from.editingFinished.connect(self._on_spin_frequency_from_edited)
        self._spin_frequency_to.editingFinished.connect(self._on_spin_frequency_to_edited)
        self._spin_frequency_center.editingFinished.connect(self._on_spin_frequency_center_edited)
        self._spin_frequency_deviation.editingFinished.connect(self._on_spin_frequency_deviation_edited)

    def load_settings(self) -> None:
        self._settings.beginGroup('search')
        self._settings.beginGroup('frequency')
        self._frequency_from = self._settings.value('from', self._spin_frequency_from.value(), float)
        self._frequency_to = self._settings.value('to', self._spin_frequency_to.value(), float)
        self._frequency_center = self._settings.value('center', self._spin_frequency_center.value(), float)
        self._frequency_deviation = self._settings.value('deviation', self._spin_frequency_deviation.value(), float)
        self._settings.endGroup()
        self._settings.endGroup()

    def save_settings(self) -> None:
        self._settings.beginGroup('search')
        self._settings.beginGroup('frequency')
        self._settings.setValue('from', self._frequency_from)
        self._settings.setValue('to', self._frequency_to)
        self._settings.setValue('center', self._frequency_center)
        self._settings.setValue('deviation', self._frequency_deviation)
        self._settings.endGroup()
        self._settings.endGroup()

    @property
    def min_frequency(self) -> float:
        if self.currentWidget() is self._page_by_range:
            return self._frequency_from
        else:
            return self._frequency_center - self._frequency_deviation

    @property
    def max_frequency(self) -> float:
        if self.currentWidget() is self._page_by_range:
            return self._frequency_to
        else:
            return self._frequency_center + self._frequency_deviation

    def set_frequency_limits(self, min_value: float, max_value: float) -> None:
        frequency_spins: list[QDoubleSpinBox] = [self._spin_frequency_from, self._spin_frequency_to,
                                                 self._spin_frequency_center]
        min_value = self._settings.from_mhz(min_value)
        max_value = self._settings.from_mhz(max_value)
        spin: QDoubleSpinBox
        for spin in frequency_spins:
            spin.setMinimum(min_value)
            spin.setMaximum(max_value)

    @Slot()
    def _on_spin_frequency_from_edited(self) -> None:
        self._frequency_from = self._settings.to_mhz(self._spin_frequency_from.value())

    @Slot()
    def _on_spin_frequency_to_edited(self) -> None:
        self._frequency_to = self._settings.to_mhz(self._spin_frequency_to.value())

    @Slot()
    def _on_spin_frequency_center_edited(self) -> None:
        self._frequency_center = self._settings.to_mhz(self._spin_frequency_center.value())

    @Slot()
    def _on_spin_frequency_deviation_edited(self) -> None:
        self._frequency_deviation = self._settings.to_mhz(self._spin_frequency_deviation.value())

    def fill_parameters(self) -> None:
        frequency_suffix: int = self._settings.frequency_unit
        frequency_suffix_str: str = ' ' + self._settings.FREQUENCY_UNITS[frequency_suffix]
        if frequency_suffix in (0, 1, 2):  # MHz, GHz, cm⁻¹
            self._spin_frequency_from.setValue(self._settings.from_mhz(self._frequency_from))
            self._spin_frequency_to.setValue(self._settings.from_mhz(self._frequency_to))
            self._spin_frequency_center.setValue(self._settings.from_mhz(self._frequency_center))
            self._spin_frequency_deviation.setValue(self._settings.from_mhz(self._frequency_deviation))
        elif frequency_suffix == 3:  # nm
            self._spin_frequency_from.setValue(mhz_to_nm(self._frequency_from))
            self._spin_frequency_to.setValue(mhz_to_nm(self._frequency_to))
            self._spin_frequency_center.setValue(mhz_to_nm(self._frequency_center))
            self._spin_frequency_deviation.setValue(
                abs(mhz_to_nm(self._frequency_center - self._frequency_deviation) -
                    mhz_to_nm(self._frequency_center + self._frequency_deviation)) / 2.0)
        else:
            raise IndexError('Wrong frequency unit index', frequency_suffix)
        precision: int = [4, 7, 8, 8][frequency_suffix]
        step_factor: float = [2.5, 2.5, 2.5, 0.25][frequency_suffix]
        frequency_spins: list[QDoubleSpinBox] = [self._spin_frequency_from, self._spin_frequency_to,
                                                 self._spin_frequency_center, self._spin_frequency_deviation]
        for spin in frequency_spins:
            spin.setSuffix(frequency_suffix_str)
            spin.setDecimals(precision)
            spin.setSingleStep(step_factor * self._spin_frequency_deviation.value())
